from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, Sequence, TypeVar, overload

from infrahub_sdk.utils import is_valid_uuid
from infrahub_sdk.uuidt import UUIDT

from infrahub.computed_attribute.jinja2 import InfrahubJinja2Template
from infrahub.core import registry
from infrahub.core.changelog.models import NodeChangelog
from infrahub.core.constants import (
    GLOBAL_BRANCH_NAME,
    OBJECT_TEMPLATE_RELATIONSHIP_NAME,
    SYSTEM_USER_ID,
    BranchSupportType,
    ComputedAttributeKind,
    InfrahubKind,
    RelationshipCardinality,
    RelationshipKind,
)
from infrahub.core.constants.schema import RESOURCE_POOL_REL_SUFFIX, SchemaElementPathType
from infrahub.core.metadata.interface import MetadataInterface
from infrahub.core.metadata.model import MetadataInfo
from infrahub.core.protocols import CoreNumberPool, CoreObjectTemplate
from infrahub.core.query.node import NodeCheckIDQuery, NodeCreateAllQuery, NodeDeleteQuery, NodeUpdateMetadataQuery
from infrahub.core.schema import (
    AttributeSchema,
    NodeSchema,
    NonGenericSchemaTypes,
    ProfileSchema,
    RelationshipSchema,
    TemplateSchema,
)
from infrahub.core.schema.attribute_parameters import NumberPoolParameters
from infrahub.core.timestamp import Timestamp
from infrahub.exceptions import InitializationError, NodeNotFoundError, PoolExhaustedError, ValidationError
from infrahub.pools.default_allocator import DefaultPoolAllocator
from infrahub.pools.noop_allocator import NoOpPoolAllocator
from infrahub.profiles.mandatory_fields_checker import ProfilesMandatoryFieldGetter
from infrahub.templates.node_applier import NodeTemplateApplier
from infrahub.types import ATTRIBUTE_TYPES

from ...graphql.constants import KIND_GRAPHQL_FIELD_NAME
from ...log import get_logger
from ..attribute import BaseAttribute
from ..query.relationship import RelationshipDeleteAllQuery
from ..relationship import RelationshipManager
from .base import BaseNode, BaseNodeMeta, BaseNodeOptions
from .node_property_attribute import DisplayLabel, HumanFriendlyIdentifier

if TYPE_CHECKING:
    from typing_extensions import Self

    from infrahub.core.branch import Branch
    from infrahub.core.creation_context import NodeCreationContext
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.core.schema.schema_branch_display import TemplateLabel
    from infrahub.core.schema.schema_branch_hfid import HFIDDefinition
    from infrahub.database import InfrahubDatabase

SchemaProtocol = TypeVar("SchemaProtocol")

# ---------------------------------------------------------------------------------------
# Type of Nodes
#  - Core node, wo/ branch : Branch, MergeRequest, Comment
#  - Core node, w/ branch : Repository, GQLQuery, Permission, Account, Groups, Schema
#  - Location Node : Location,
#  - Select Node : Status, Role, Manufacturer etc ..
#  -
# ---------------------------------------------------------------------------------------

log = get_logger()


class Node(BaseNode, MetadataInterface, metaclass=BaseNodeMeta):
    @classmethod
    def __init_subclass_with_meta__(
        cls, _meta: BaseNodeOptions | None = None, default_filter: None = None, **options: dict[str, Any]
    ) -> None:
        if not _meta:
            _meta = BaseNodeOptions(cls)

        _meta.default_filter = default_filter
        super().__init_subclass_with_meta__(_meta=_meta, **options)

    def __init__(self, schema: NodeSchema | ProfileSchema | TemplateSchema, branch: Branch, at: Timestamp) -> None:
        super().__init__()
        self._schema: NodeSchema | ProfileSchema | TemplateSchema = schema
        self._branch: Branch = branch
        self._at: Timestamp = at
        self._existing: bool = False
        self._metadata = MetadataInfo()

        self.id: str = None
        self.db_id: str = None

        self._source: Node | None = None
        self._owner: Node | None = None
        self._is_protected: bool = None
        self._computed_jinja2_attributes: list[str] = []
        self._profile_provided_attrs: set[str] = set()
        self._profile_provided_rels: set[str] = set()

        self._display_label: DisplayLabel | None = None
        self._human_friendly_id: HumanFriendlyIdentifier | None = None

        # Lists of attributes and relationships names
        self._attributes: list[str] = []
        self._relationships: list[str] = []
        self._node_changelog: NodeChangelog | None = None
        self._creation_context: NodeCreationContext | None = None

    def _set_created_at(self, value: Timestamp | None) -> None:
        self._metadata.created_at = value

    def _set_created_by(self, value: str | None) -> None:
        self._metadata.created_by = value

    def _set_updated_at(self, value: Timestamp | None) -> None:
        self._metadata.updated_at = value

    def _set_updated_by(self, value: str | None) -> None:
        self._metadata.updated_by = value

    def _get_created_at(self) -> Timestamp | None:
        return self._metadata.created_at

    def _get_created_by(self) -> str | None:
        return self._metadata.created_by

    def _get_updated_at(self) -> Timestamp | None:
        return self._metadata.updated_at

    def _get_updated_by(self) -> str | None:
        return self._metadata.updated_by

    def get_schema(self) -> NonGenericSchemaTypes:
        return self._schema

    def get_branch(self) -> Branch:
        return self._branch

    def get_kind(self) -> str:
        """Return the main Kind of the Object."""
        return self._schema.kind

    def get_id(self) -> str:
        """Return the ID of the node"""
        if self.id:
            return self.id

        raise InitializationError("The node has not been saved yet and doesn't have an id")

    def get_attribute(self, name: str) -> BaseAttribute:
        attribute = getattr(self, name, None)
        if not isinstance(attribute, BaseAttribute):
            raise ValueError(f"{name} is not an attribute of {self.get_kind()}")
        return attribute

    def get_relationship(self, name: str) -> RelationshipManager:
        relationship = getattr(self, name, None)
        if not isinstance(relationship, RelationshipManager):
            raise ValueError(f"{name} is not a relationship of {self.get_kind()}")
        return relationship

    def get_relationship_by_identifier(self, identifier: str) -> RelationshipManager:
        for rel_schema in self._schema.relationships:
            if rel_schema.identifier == identifier:
                return self.get_relationship(rel_schema.name)
        raise ValueError(f"Unable to find the relationship with the identifier {identifier} for {self.get_kind()}")

    def uses_profiles(self) -> bool:
        for attr_name in self.get_schema().attribute_names:
            try:
                node_attr = self.get_attribute(attr_name)
            except ValueError:
                continue
            if node_attr and node_attr.is_from_profile:
                return True
        return False

    async def get_hfid(self, db: InfrahubDatabase, include_kind: bool = False) -> list[str] | None:
        """Return the Human friendly id of the node."""
        if not self._schema.human_friendly_id:
            return None

        hfid_values: list[str] | None = None
        if self._human_friendly_id:
            hfid_values = self._human_friendly_id.get_value(node=self, at=self._at)
        if not hfid_values:
            hfid_values = [await self.get_path_value(db=db, path=item) for item in self._schema.human_friendly_id]

        hfid = [value for value in hfid_values if value is not None]
        return [self.get_kind()] + hfid if include_kind else hfid

    async def get_hfid_as_string(self, db: InfrahubDatabase, include_kind: bool = False) -> str | None:
        """Return the Human friendly id of the node in string format separated with a dunder (__) ."""
        hfid = await self.get_hfid(db=db, include_kind=include_kind)
        if not hfid:
            return None
        return "__".join(hfid)

    def has_human_friendly_id(self) -> bool:
        return self._human_friendly_id is not None

    async def add_human_friendly_id(self, db: InfrahubDatabase) -> None:
        if self._human_friendly_id:
            return

        self._human_friendly_id = HumanFriendlyIdentifier(
            node_schema=self._schema, template=self._schema.human_friendly_id
        )
        await self._human_friendly_id.compute(db=db, node=self)

    async def get_display_label(self, db: InfrahubDatabase) -> str:
        if self._display_label and (value := self._display_label.get_value(node=self, at=self._at)):
            return value

        return await self.render_display_label(db=db)

    def has_display_label(self) -> bool:
        return self._display_label is not None

    async def add_display_label(self, db: InfrahubDatabase) -> None:
        if self._display_label:
            return

        self._display_label = DisplayLabel(node_schema=self._schema, template=self._schema.display_label)
        await self._display_label.compute(db=db, node=self)

    async def get_path_value(self, db: InfrahubDatabase, path: str) -> str:
        schema_path = self._schema.parse_schema_path(
            path=path, schema=db.schema.get_schema_branch(name=self._branch.name)
        )

        if not schema_path.has_property:
            raise ValueError(f"Unable to retrieve the value of a path without property {path!r} on {self.get_kind()!r}")

        if (
            schema_path.is_type_relationship
            and schema_path.relationship_schema.cardinality == RelationshipCardinality.MANY
        ):
            raise ValueError(
                f"Unable to retrieve the value of a path on a relationship of cardinality many {path!r} on {self.get_kind()!r}"
            )

        if schema_path.is_type_attribute:
            attr = getattr(self, schema_path.attribute_schema.name)
            return getattr(attr, schema_path.attribute_property_name)

        if schema_path.is_type_relationship:
            relm: RelationshipManager = getattr(self, schema_path.relationship_schema.name)
            await relm.resolve(db=db)
            node = await relm.get_peer(db=db)
            attr = getattr(node, schema_path.attribute_schema.name)
            return getattr(attr, schema_path.attribute_property_name)

    def get_labels(self) -> list[str]:
        """Return the labels for this object, composed of the kind
        and the list of Generic this object is inheriting from."""
        labels: list[str] = []
        if isinstance(self._schema, NodeSchema):
            labels = [self.get_kind()] + self._schema.inherit_from
            if (
                self._schema.namespace not in ["Schema", "Internal"]
                and InfrahubKind.GENERICGROUP not in self._schema.inherit_from
            ):
                labels.append(InfrahubKind.NODE)
            return labels

        if isinstance(self._schema, ProfileSchema | TemplateSchema):
            labels = [self.get_kind()] + self._schema.inherit_from
            return labels

        return [self.get_kind()]

    def get_branch_based_on_support_type(self) -> Branch:
        """If the attribute is branch aware, return the Branch object associated with this attribute
        If the attribute is branch agnostic return the Global Branch

        Returns:
            Branch:
        """
        if self._schema.branch == BranchSupportType.AGNOSTIC:
            return registry.get_global_branch()
        return self._branch

    def __repr__(self) -> str:
        v = f"{self.get_kind()}(ID: {str(self.id)})"
        return v if self._existing else f"{v}[NEW]"

    @property
    def has_changelog(self) -> bool:
        return self._node_changelog is not None

    @property
    def node_changelog(self) -> NodeChangelog:
        if self._node_changelog:
            return self._node_changelog

        raise InitializationError("The node has not been saved so no changelog exists")

    @overload
    @classmethod
    async def init(
        cls,
        schema: NodeSchema | ProfileSchema | TemplateSchema | str,
        db: InfrahubDatabase,
        branch: Branch | str | None = ...,
        at: Timestamp | str | None = ...,
    ) -> Self: ...

    @overload
    @classmethod
    async def init(
        cls,
        schema: type[SchemaProtocol],
        db: InfrahubDatabase,
        branch: Branch | str | None = ...,
        at: Timestamp | str | None = ...,
    ) -> SchemaProtocol: ...

    @classmethod
    async def init(
        cls,
        schema: NodeSchema | ProfileSchema | TemplateSchema | str | type[SchemaProtocol],
        db: InfrahubDatabase,
        branch: Branch | str | None = None,
        at: Timestamp | str | None = None,
    ) -> Self | SchemaProtocol:
        attrs: dict[str, Any] = {}

        branch = await registry.get_branch(branch=branch, db=db)

        if isinstance(schema, NodeSchema | ProfileSchema | TemplateSchema):
            attrs["schema"] = schema
        elif isinstance(schema, str):
            # TODO need to raise a proper exception for this, right now it will raise a generic ValueError
            attrs["schema"] = db.schema.get(name=schema, branch=branch)
        elif hasattr(schema, "_is_runtime_protocol") and schema._is_runtime_protocol:
            attrs["schema"] = db.schema.get(name=schema.__name__, branch=branch)
        else:
            raise ValueError(
                f"Invalid schema provided {type(schema)}, expected NodeSchema, ProfileSchema or TemplateSchema"
            )

        attrs["branch"] = branch
        attrs["at"] = Timestamp(at)

        return cls(**attrs)

    async def handle_pool(
        self,
        db: InfrahubDatabase,
        attribute: BaseAttribute,
        allocate_resources: bool = True,
    ) -> None:
        """Evaluate if a resource has been requested from a pool and apply the resource

        This method only works on number pools, currently Integer is the only type that has the from_pool
        within the create code.

        Supports two cases:
        1. Schema-defined NumberPool attributes (kind="NumberPool" with number_pool_id in parameters)
        2. User-specified from_pool (user explicitly passes {"from_pool": {"id": pool_id}} to a Number attribute)
        """
        number_pool_id: str | None = None
        # Templates must use _from_resource_pool relationships, not from_pool
        if isinstance(self._schema, TemplateSchema):
            if attribute.from_pool:
                pool_rel_name = f"{attribute.name}{RESOURCE_POOL_REL_SUFFIX}"
                raise ValidationError(
                    {
                        f"{attribute.name}.from_pool": (
                            f"'from_pool' is not supported on template attributes. Set the '{pool_rel_name}' relationship on this template instead."
                        )
                    }
                )

        # Case 1: Schema-defined NumberPool attribute
        if (
            not number_pool_id
            and attribute.schema.kind == "NumberPool"
            and isinstance(attribute.schema.parameters, NumberPoolParameters)
        ):
            number_pool_id = attribute.schema.parameters.number_pool_id
            if not number_pool_id:
                raise ValidationError(
                    {f"{attribute.name}": f"The pool for {attribute.name} has not been provisioned yet."}
                )
            attribute.from_pool = {"id": number_pool_id}
            attribute.is_default = False
        # Case 2: User-specified from_pool on a regular Number attribute
        elif not number_pool_id and attribute.from_pool:
            number_pool_id = attribute.from_pool.get("id")
            if not number_pool_id:
                raise ValidationError({f"{attribute.name}.from_pool": "No pool ID specified in from_pool."})

        if not number_pool_id:
            # no pool allocation necessary
            return

        try:
            if is_valid_uuid(number_pool_id):
                number_pool = await registry.manager.get_one(
                    db=db, id=number_pool_id, kind=CoreNumberPool, raise_on_error=True
                )
            else:
                results = await registry.manager.query(
                    db=db, schema=InfrahubKind.NUMBERPOOL, filters={"name__value": number_pool_id}
                )
                if not results:
                    raise NodeNotFoundError(node_type=InfrahubKind.NUMBERPOOL, identifier=number_pool_id)
                number_pool = results[0]
        except NodeNotFoundError as exc:
            raise ValidationError(
                {f"{attribute.name}.from_pool": f"The pool requested {attribute.from_pool} was not found."}
            ) from exc

        if not allocate_resources:
            attribute.source = number_pool.id
            return

        if (
            number_pool.node.value in [self._schema.kind] + self._schema.inherit_from
            and number_pool.node_attribute.value == attribute.name
        ):
            try:
                next_free = await number_pool.get_resource(
                    db=db, branch=self._branch, identifier=self.get_id(), attribute=attribute.schema
                )
            except PoolExhaustedError as exc:
                raise ValidationError(
                    {f"{attribute.name}.from_pool": f"The pool {number_pool.node.value} is exhausted."}
                ) from exc

            attribute.value = next_free
            attribute.source = number_pool.id
        else:
            raise ValidationError(
                {
                    f"{attribute.name}.from_pool": f"The {number_pool.name.value} pool can't be used for '{attribute.name}'."
                }
            )

    async def handle_object_template(
        self, fields: dict, db: InfrahubDatabase, errors: list, process_pools: bool = True
    ) -> None:
        """Fill the `fields` parameters with values from an object template if one is in use."""
        object_template_field = fields.get(OBJECT_TEMPLATE_RELATIONSHIP_NAME)
        if not object_template_field:
            return

        try:
            template: CoreObjectTemplate = await registry.manager.find_object(
                db=db,
                kind=self._schema.get_relationship(name=OBJECT_TEMPLATE_RELATIONSHIP_NAME).peer,
                id=object_template_field.get("id"),
                hfid=object_template_field.get("hfid"),
                branch=self.get_branch_based_on_support_type(),
            )
        except NodeNotFoundError:
            errors.append(
                ValidationError(
                    {
                        f"{OBJECT_TEMPLATE_RELATIONSHIP_NAME}": (
                            "Unable to find the object template in the database "
                            f"'{object_template_field.get('id') or object_template_field.get('hfid')}'"
                        )
                    }
                )
            )
            return

        pool_allocator = DefaultPoolAllocator(db=db, branch=self._branch) if process_pools else NoOpPoolAllocator()
        applier = NodeTemplateApplier(db=db, branch=self._branch, pool_allocator=pool_allocator)
        applied_fields = await applier.apply(
            template=template, target_schema=self._schema, target_id=self.id, user_fields=fields
        )

        # Update fields dict in-place with applied values
        # Only add new keys, don't overwrite existing ones (user fields take precedence)
        for key, value in applied_fields.items():
            if key not in fields:
                fields[key] = value

    async def _get_profile_provided_mandatory_fields(
        self, db: InfrahubDatabase, fields: dict[str, Any]
    ) -> tuple[set[str], set[str]]:
        if not isinstance(self._schema, NodeSchema) or "profiles" not in fields:
            return set(), set()

        mandatory_attrs_to_check = [a for a in self._schema.mandatory_attribute_names if a not in fields.keys()]
        mandatory_rels_to_check = [r for r in self._schema.mandatory_relationship_names if r not in fields.keys()]
        if not mandatory_attrs_to_check and not mandatory_rels_to_check:
            return set(), set()

        profiles_mandatory_field_getter = ProfilesMandatoryFieldGetter(db=db, branch=self._branch)
        return await profiles_mandatory_field_getter.get_mandatory_fields_from_profiles(
            schema=self._schema,
            profiles_data=fields.get("profiles"),
            mandatory_attr_names=mandatory_attrs_to_check,
            mandatory_rel_names=mandatory_rels_to_check,
        )

    async def _process_fields(self, fields: dict, db: InfrahubDatabase, process_pools: bool = True) -> None:
        errors = []

        if "_source" in fields.keys():
            self._source = fields["_source"]
        if "_owner" in fields.keys():
            self._owner = fields["_owner"]

        # -------------------------------------------
        # Validate Input
        # -------------------------------------------
        if "updated_at" in fields and "updated_at" not in self._schema.valid_input_names:
            # FIXME: Allow users to use "updated_at" named attributes until we have proper metadata handling
            fields.pop("updated_at")
        for field_name in fields.keys():
            if field_name not in self._schema.valid_input_names:
                log.error(f"{field_name} is not a valid input for {self.get_kind()}")

        # Backfill fields with the ones from the template if there's one
        await self.handle_object_template(fields=fields, db=db, errors=errors, process_pools=process_pools)

        if not self._existing:
            (
                self._profile_provided_attrs,
                self._profile_provided_rels,
            ) = await self._get_profile_provided_mandatory_fields(db=db, fields=fields)

            for mandatory_attr in self._schema.mandatory_attribute_names:
                if mandatory_attr not in fields.keys() and mandatory_attr not in self._profile_provided_attrs:
                    if self._schema.is_node_schema:
                        mandatory_attribute = self._schema.get_attribute(name=mandatory_attr)
                        if (
                            mandatory_attribute.computed_attribute
                            and mandatory_attribute.computed_attribute.kind == ComputedAttributeKind.JINJA2
                        ):
                            self._computed_jinja2_attributes.append(mandatory_attr)
                            continue

                        if mandatory_attribute.kind == "NumberPool":
                            continue

                    errors.append(
                        ValidationError({mandatory_attr: f"{mandatory_attr} is mandatory for {self.get_kind()}"})
                    )

            for mandatory_rel in self._schema.mandatory_relationship_names:
                if mandatory_rel not in fields.keys() and mandatory_rel not in self._profile_provided_rels:
                    errors.append(
                        ValidationError({mandatory_rel: f"{mandatory_rel} is mandatory for {self.get_kind()}"})
                    )

        if errors:
            raise ValidationError(errors)

        # -------------------------------------------
        # Generate Attribute and Relationship and assign them
        # -------------------------------------------
        errors.extend(await self._process_fields_relationships(fields=fields, db=db))
        errors.extend(await self._process_fields_attributes(fields=fields, db=db, process_pools=process_pools))

        if errors:
            raise ValidationError(errors)

        # Check if any post processor have been defined
        # A processor can be used for example to assigne a default value
        for name in self._attributes + self._relationships:
            if hasattr(self, f"process_{name}"):
                await getattr(self, f"process_{name}")(db=db)

    async def _process_fields_relationships(self, fields: dict, db: InfrahubDatabase) -> list[ValidationError]:
        errors: list[ValidationError] = []

        for rel_schema in self._schema.relationships:
            self._relationships.append(rel_schema.name)

            # Check if there is a more specific generator present
            # Otherwise use the default generator
            generator_method_name = "_generate_relationship_default"
            if hasattr(self, f"generate_{rel_schema.name}"):
                generator_method_name = f"generate_{rel_schema.name}"

            generator_method = getattr(self, generator_method_name)
            try:
                setattr(
                    self,
                    rel_schema.name,
                    await generator_method(
                        db=db, name=rel_schema.name, schema=rel_schema, data=fields.get(rel_schema.name)
                    ),
                )
            except ValidationError as exc:
                errors.append(exc)

        return errors

    async def _process_fields_attributes(
        self, fields: dict, db: InfrahubDatabase, process_pools: bool
    ) -> list[ValidationError]:
        errors: list[ValidationError] = []

        for attr_schema in self._schema.attributes:
            self._attributes.append(attr_schema.name)
            if not self._existing and attr_schema.name in self._computed_jinja2_attributes:
                continue

            # Check if there is a more specific generator present
            # Otherwise use the default generator
            generator_method_name = "_generate_attribute_default"
            if hasattr(self, f"generate_{attr_schema.name}"):
                generator_method_name = f"generate_{attr_schema.name}"

            generator_method = getattr(self, generator_method_name)
            try:
                setattr(
                    self,
                    attr_schema.name,
                    await generator_method(
                        db=db, name=attr_schema.name, schema=attr_schema, data=fields.get(attr_schema.name)
                    ),
                )
                if not self._existing:
                    attribute: BaseAttribute = getattr(self, attr_schema.name)
                    await self.handle_pool(db=db, attribute=attribute, allocate_resources=process_pools)

                    if attr_schema.name in self._profile_provided_attrs:
                        continue

                    if process_pools or attribute.from_pool is None:
                        attribute.validate(value=attribute.value, name=attribute.name, schema=attribute.schema)
            except ValidationError as exc:
                errors.append(exc)

        return errors

    async def _process_macros(self, db: InfrahubDatabase) -> None:
        schema_branch = db.schema.get_schema_branch(self._branch.name)
        allowed_path_types = (
            SchemaElementPathType.ATTR_WITH_PROP
            | SchemaElementPathType.REL_ONE_MANDATORY_ATTR_WITH_PROP
            | SchemaElementPathType.REL_ONE_OPTIONAL_ATTR_WITH_PROP
        )
        errors = []
        for macro in self._computed_jinja2_attributes:
            variables = {}
            attr_schema = self._schema.get_attribute(name=macro)
            if not attr_schema.computed_attribute:
                errors.append(
                    ValidationError({macro: f"{macro} is missing computational_logic for macro ({attr_schema.kind})"})
                )
                continue
            if not attr_schema.computed_attribute.jinja2_template:
                errors.append(
                    ValidationError({macro: f"{macro} is missing computational_logic for macro ({attr_schema.kind})"})
                )
                continue

            jinja_template = InfrahubJinja2Template(template=attr_schema.computed_attribute.jinja2_template)
            for variable in jinja_template.get_variables():
                attribute_path = schema_branch.validate_schema_path(
                    node_schema=self._schema, path=variable, allowed_path_types=allowed_path_types
                )
                if attribute_path.is_type_relationship:
                    relationship_attribute: RelationshipManager = getattr(
                        self, attribute_path.active_relationship_schema.name
                    )
                    if peer := await relationship_attribute.get_peer(db=db, raise_on_error=False):
                        related_node = await registry.manager.get_one_by_id_or_default_filter(
                            db=db,
                            id=peer.id,
                            kind=attribute_path.active_relationship_schema.peer,
                            branch=self._branch.name,
                        )

                        attribute: BaseAttribute = getattr(
                            getattr(related_node, attribute_path.active_attribute_schema.name),
                            attribute_path.active_attribute_property_name,
                        )
                        variables[variable] = attribute
                    else:
                        variables[variable] = None

                elif attribute_path.is_type_attribute:
                    attribute = getattr(
                        getattr(self, attribute_path.active_attribute_schema.name),
                        attribute_path.active_attribute_property_name,
                    )
                    variables[variable] = attribute

            content = await jinja_template.render(variables=variables)

            generator_method_name = "_generate_attribute_default"
            if hasattr(self, f"generate_{attr_schema.name}"):
                generator_method_name = f"generate_{attr_schema.name}"

            generator_method = getattr(self, generator_method_name)
            try:
                setattr(
                    self,
                    attr_schema.name,
                    await generator_method(db=db, name=attr_schema.name, schema=attr_schema, data=content),
                )
                attribute = getattr(self, attr_schema.name)

                attribute.validate(value=attribute.value, name=attribute.name, schema=attribute.schema)
            except ValidationError as exc:
                errors.append(exc)

        if errors:
            raise ValidationError(errors)

    async def _generate_relationship_default(
        self,
        name: str,  # noqa: ARG002
        schema: RelationshipSchema,
        data: Any,
        db: InfrahubDatabase,
    ) -> RelationshipManager:
        rm = await RelationshipManager.init(
            db=db,
            data=data,
            schema=schema,
            branch=self._branch,
            at=self._at,
            node=self,
        )

        return rm

    async def _generate_attribute_default(
        self,
        name: str,
        schema: AttributeSchema,
        data: Any,
        db: InfrahubDatabase,  # noqa: ARG002
    ) -> BaseAttribute:
        attr_class = ATTRIBUTE_TYPES[schema.kind].get_infrahub_class()
        attr = attr_class(
            data=data,
            name=name,
            schema=schema,
            branch=self._branch,
            at=self._at,
            node=self,
            source=self._source,
            owner=self._owner,
        )
        if isinstance(data, dict):
            attr._set_created_at(data.pop("created_at", None))
            attr._set_created_by(data.pop("created_by", None))
            attr._set_updated_at(data.pop("updated_at", None))
            attr._set_updated_by(data.pop("updated_by", None))

        return attr

    async def process_label(self, db: InfrahubDatabase | None = None) -> None:  # noqa: ARG002
        # If there label and name are both defined for this node
        #  if label is not define, we'll automatically populate it with a human friendy vesion of name
        if not self._existing and hasattr(self, "label") and hasattr(self, "name"):
            if self.label.value is None and self.name.value:
                self.label.value = " ".join([word.title() for word in self.name.value.split("_")])
                self.label.is_default = False

    async def new(self, db: InfrahubDatabase, id: str | None = None, process_pools: bool = True, **kwargs: Any) -> Self:
        if id and not is_valid_uuid(id):
            raise ValidationError({"id": f"{id} is not a valid UUID"})
        if id:
            query = await NodeCheckIDQuery.init(db=db, node_id=id)
            if await query.count(db=db):
                raise ValidationError({"id": f"{id} is already in use"})

        self.id = id or str(UUIDT())

        await self._process_fields(db=db, fields=kwargs, process_pools=process_pools)
        await self._process_macros(db=db)

        return self

    @staticmethod
    def _merge_relationship_fields(
        definitions: list[HFIDDefinition | TemplateLabel],
    ) -> dict[str, set[str]]:
        """Merge ``relationship_fields`` from the given definitions into a single mapping."""
        extra_filters: dict[str, set[str]] = {}
        for definition in definitions:
            for rel_name, attrs in definition.relationship_fields.items():
                extra_filters.setdefault(rel_name, set()).update(attrs)
        return extra_filters

    def _collect_extra_filters(self, schema_branch: SchemaBranch) -> dict[str, set[str]]:
        """Collect peer attributes that must be loaded during relationship resolution."""
        definitions: list[HFIDDefinition | TemplateLabel] = []

        # If we are creating a new node, we need to resolve extra filters from Display Labels or HFIDs, if we don't do
        # this the fields might be blank.
        # We could also need it when we need to recompute the Display Labels or HFIDs
        if (not self._existing) or self._human_friendly_id:
            if hfid := schema_branch.hfids.get_template_nodes().get(self._schema.kind):
                definitions.append(hfid)
        if (not self._existing) or self._display_label:
            if display_labels := schema_branch.display_labels.get_template_nodes().get(self._schema.kind):
                definitions.append(display_labels)

        return self._merge_relationship_fields(definitions)

    async def resolve_relationships(self, db: InfrahubDatabase, user_id: str = SYSTEM_USER_ID) -> None:
        schema_branch = db.schema.get_schema_branch(name=self.get_branch_based_on_support_type().name)
        extra_filters = self._collect_extra_filters(schema_branch=schema_branch)

        for name in self._relationships:
            relm: RelationshipManager = getattr(self, name)
            query_filter = []
            if name in extra_filters:
                query_filter.extend(list(extra_filters[name]))

            await relm.resolve(db=db, fields=query_filter, user_id=user_id)

    async def load(
        self,
        db: InfrahubDatabase,
        id: str | None = None,
        db_id: str | None = None,
        **kwargs: Any,
    ) -> Self:
        self.id = id
        self.db_id = db_id
        self._existing = True

        if not self._schema.is_schema_node:
            if hfid := kwargs.pop("human_friendly_id", None):
                self._human_friendly_id = HumanFriendlyIdentifier(
                    node_schema=self._schema, template=self._schema.human_friendly_id, value=hfid
                )
            if display_label := kwargs.pop("display_label", None):
                self._display_label = DisplayLabel(
                    node_schema=self._schema, template=self._schema.display_label, value=display_label
                )

        await self._process_fields(db=db, fields=kwargs)
        return self

    async def _create(self, db: InfrahubDatabase, user_id: str, at: Timestamp | None = None) -> NodeChangelog:
        create_at = Timestamp(at)
        self._set_created_at(create_at)
        self._set_created_by(user_id)
        self._set_updated_at(create_at)
        self._set_updated_by(user_id)

        if not self._schema.is_schema_node:
            await self.add_human_friendly_id(db=db)
            await self.add_display_label(db=db)

        query = await NodeCreateAllQuery.init(db=db, node=self, at=create_at, user_id=user_id)
        await query.execute(db=db)

        _, self.db_id = query.get_self_ids()
        self._at = create_at
        self._existing = True

        new_ids = query.get_ids()
        node_changelog = NodeChangelog(node_id=self.get_id(), node_kind=self.get_kind(), display_label="")

        if self._human_friendly_id:
            node_changelog.create_attribute(
                attribute=self._human_friendly_id.get_node_attribute(node=self, at=create_at)
            )
        if self._display_label:
            node_changelog.create_attribute(attribute=self._display_label.get_node_attribute(node=self, at=create_at))

        # Go over the list of Attribute and assign the new IDs one by one
        for name in self._attributes:
            attr: BaseAttribute = getattr(self, name)
            attr.id, attr.db_id = new_ids[name]
            attr.at = create_at
            attr._set_created_at(create_at)
            attr._set_created_by(user_id)
            attr._set_updated_at(create_at)
            attr._set_updated_by(user_id)
            node_changelog.create_attribute(attribute=attr)

        # Go over the list of relationships and assign the new IDs one by one
        for name in self._relationships:
            relm: RelationshipManager = getattr(self, name)
            for rel in relm._relationships:
                identifier = f"{rel.schema.identifier}::{rel.peer_id}"
                rel.id, rel.db_id = new_ids[identifier]
                node_changelog.create_relationship(relationship=rel)

        node_changelog.display_label = await self.get_display_label(db=db)
        return node_changelog

    async def _update(
        self, db: InfrahubDatabase, user_id: str, at: Timestamp | None = None, fields: list[str] | None = None
    ) -> NodeChangelog:
        """Update the node in the database if needed."""

        update_at = Timestamp(at)
        node_changelog = NodeChangelog(node_id=self.get_id(), node_kind=self.get_kind(), display_label="")

        # Go over the list of Attribute and update them one by one
        for name in self._attributes:
            if (fields and name in fields) or not fields:
                attr = self.get_attribute(name=name)
                updated_attribute = await attr.save(db=db, user_id=user_id, at=update_at)
                if updated_attribute:
                    node_changelog.add_attribute(attribute=updated_attribute)

        # Go over the list of relationships and update them one by one
        processed_relationships: list[str] = []
        for name in self._relationships:
            if (fields and name in fields) or not fields:
                processed_relationships.append(name)
                rel = self.get_relationship(name=name)
                updated_relationship = await rel.save(db=db, user_id=user_id, at=update_at)
                node_changelog.add_relationship(relationship_changelog=updated_relationship)

        if len(processed_relationships) != len(self._relationships):
            # Analyze if the node has a parent and add it to the changelog if missing
            if parent_relationship := self._get_parent_relationship_name():
                if parent_relationship not in processed_relationships:
                    rel = self.get_relationship(name=parent_relationship)
                    if parent := await rel.get_parent(db=db):
                        node_changelog.add_parent_from_relationship(parent=parent)

        # Update the HFID if one of its variables is being updated
        if self._human_friendly_id and (
            (fields and "human_friendly_id" in fields) or self._human_friendly_id.needs_update(fields=fields)
        ):
            await self._human_friendly_id.compute(db=db, node=self)
            updated_attribute = await self._human_friendly_id.get_node_attribute(node=self, at=update_at).save(
                at=update_at, db=db
            )
            if updated_attribute:
                node_changelog.add_attribute(attribute=updated_attribute)

        # Update the display label if one of its variables is being updated
        if self._display_label and (
            (fields and "display_label" in fields) or self._display_label.needs_update(fields=fields)
        ):
            await self._display_label.compute(db=db, node=self)
            self._display_label.get_node_attribute(node=self, at=update_at).get_create_data(node_schema=self._schema)
            updated_attribute = await self._display_label.get_node_attribute(node=self, at=update_at).save(
                at=update_at, db=db
            )
            if updated_attribute:
                node_changelog.add_attribute(attribute=updated_attribute)

        node_changelog.display_label = await self.get_display_label(db=db)

        if node_changelog.has_changes:
            self._set_updated_at(update_at)
            self._set_updated_by(user_id)
            update_branch = self.get_branch_based_on_support_type()
            if update_branch.is_default or update_branch.is_global:
                await self._save_metadata(db=db, branch=update_branch)
        return node_changelog

    async def save(
        self,
        db: InfrahubDatabase,
        user_id: str = SYSTEM_USER_ID,
        at: Timestamp | None = None,
        fields: list[str] | None = None,
    ) -> Self:
        """Create or Update the Node in the database."""
        save_at = Timestamp(at)

        if self._existing:
            self._node_changelog = await self._update(db=db, user_id=user_id, at=save_at, fields=fields)
        else:
            self._node_changelog = await self._create(db=db, user_id=user_id, at=save_at)

        return self

    async def _save_metadata(self, db: InfrahubDatabase, branch: Branch) -> None:
        if user_id := self._get_updated_by():
            update_metadata_query = await NodeUpdateMetadataQuery.init(
                db=db, branch=branch, node_id=self.get_id(), user_id=user_id, at=self._get_updated_at()
            )
            await update_metadata_query.execute(db=db)

    async def delete(self, db: InfrahubDatabase, user_id: str = SYSTEM_USER_ID, at: Timestamp | None = None) -> None:
        """Delete the Node in the database."""

        delete_at = Timestamp(at)

        node_changelog = NodeChangelog(
            node_id=self.get_id(), node_kind=self.get_kind(), display_label=await self.get_display_label(db=db)
        )
        # Go over the list of Attribute and update them one by one
        for name in self._attributes:
            attr: BaseAttribute = getattr(self, name)
            if deleted_attribute := await attr.delete(db=db, at=delete_at, user_id=user_id):
                node_changelog.add_attribute(attribute=deleted_attribute)

        if self._human_friendly_id:
            if deleted_attribute := await self._human_friendly_id.get_node_attribute(node=self, at=delete_at).delete(
                db=db,
                user_id=user_id,
                at=delete_at,
            ):
                node_changelog.add_attribute(attribute=deleted_attribute)

        if self._display_label:
            if deleted_attribute := await self._display_label.get_node_attribute(node=self, at=delete_at).delete(
                db=db, at=delete_at, user_id=user_id
            ):
                node_changelog.add_attribute(attribute=deleted_attribute)

        branch = self.get_branch_based_on_support_type()

        delete_query = await RelationshipDeleteAllQuery.init(
            db=db,
            node_id=self.get_id(),
            branch=branch,
            user_id=user_id,
            at=delete_at,
            branch_agnostic=branch.name == GLOBAL_BRANCH_NAME,
        )
        await delete_query.execute(db=db)

        deleted_relationships_changelogs = delete_query.get_deleted_relationships_changelog(self._schema)
        for relationship_changelog in deleted_relationships_changelogs:
            node_changelog.add_relationship(relationship_changelog=relationship_changelog)

        query = await NodeDeleteQuery.init(db=db, node=self, at=delete_at, user_id=user_id)
        await query.execute(db=db)

        self._node_changelog = node_changelog

    async def to_graphql(
        self,
        db: InfrahubDatabase,
        fields: dict | None = None,
        related_node_ids: set | None = None,
        filter_sensitive: bool = False,
        permissions: dict | None = None,
        include_properties: bool = True,
    ) -> dict:
        """Generate GraphQL Payload for all attributes

        Returns:
            (dict): Return GraphQL Payload
        """

        response: dict[str, Any] = {"id": self.id, KIND_GRAPHQL_FIELD_NAME: self.get_kind()}

        if related_node_ids is not None:
            related_node_ids.add(self.id)

        FIELD_NAME_TO_EXCLUDE = ["id"] + self._schema.relationship_names

        field_names = (
            [field_name for field_name in fields.keys() if field_name not in FIELD_NAME_TO_EXCLUDE]
            if fields and isinstance(fields, dict)
            else self._schema.attribute_names + ["__typename", "display_label"]
        )

        for field_name in field_names:
            if field_name == "__typename":
                # Note we already store kind within KIND_GRAPHQL_FIELD_NAME.
                response[field_name] = self.get_kind()
                continue

            if field_name == "display_label":
                response[field_name] = await self.get_display_label(db=db)
                continue

            if field_name == "hfid":
                response[field_name] = await self.get_hfid(db=db)
                continue

            if field_name == "_updated_at":
                if updated_at := self._get_updated_at():
                    response[field_name] = await updated_at.to_graphql()
                else:
                    response[field_name] = None
                continue

            field: BaseAttribute | None = getattr(self, field_name, None)

            if not field:
                response[field_name] = None
                continue

            if fields and isinstance(fields, dict):
                response[field_name] = await field.to_graphql(
                    db=db,
                    fields=fields.get(field_name),
                    related_node_ids=related_node_ids,
                    filter_sensitive=filter_sensitive,
                    permissions=permissions,
                    include_properties=include_properties,
                )
            else:
                response[field_name] = await field.to_graphql(
                    db=db,
                    filter_sensitive=filter_sensitive,
                    permissions=permissions,
                    include_properties=include_properties,
                )

        for relationship_schema in self.get_schema().relationships:
            peer_rels = []
            if not fields or relationship_schema.name not in fields:
                continue
            rel_manager = getattr(self, relationship_schema.name, None)
            if rel_manager is None:
                continue
            try:
                if relationship_schema.cardinality is RelationshipCardinality.ONE:
                    rel = rel_manager.get_one()
                    if rel:
                        peer_rels = [rel]
                else:
                    peer_rels = list(rel_manager)
                if peer_rels:
                    response[relationship_schema.name] = [
                        {"node": {"id": relationship.peer_id}} for relationship in peer_rels if relationship.peer_id
                    ]
            except LookupError:
                continue

        return response

    async def _build_meta_response(self, field_name: str, fields: dict) -> dict:
        data = {}
        for meta_field in fields.get(field_name, {}).keys():
            if meta_field == "created_at":
                created_at = self._get_created_at()
                data["created_at"] = created_at.to_datetime() if created_at else None

            if meta_field == "created_by":
                data["created_by"] = (
                    {"id": self._get_created_by(), "__kind__": "CoreAccount"} if self._get_created_by() else None
                )

            if meta_field == "updated_by":
                data["updated_by"] = (
                    {"id": self._get_updated_by(), "__kind__": "CoreAccount"} if self._get_updated_by() else None
                )

            if meta_field == "updated_at":
                updated_at = self._get_updated_at()
                data["updated_at"] = updated_at.to_datetime() if updated_at else None
        return data

    async def from_graphql(self, data: dict, db: InfrahubDatabase, process_pools: bool = True) -> bool:
        """Update object from a GraphQL payload."""

        changed = False

        for key, value in data.items():
            if key in self._attributes and isinstance(value, dict):
                attribute = getattr(self, key)
                changed |= await attribute.from_graphql(data=value, db=db, process_pools=process_pools)

            if key in self._relationships:
                rel: RelationshipManager = getattr(self, key)
                changed |= await rel.update(db=db, data=value, process_delete=process_pools)

        return changed

    async def render_display_label(self, db: InfrahubDatabase | None = None) -> str:  # noqa: ARG002
        if not self._schema.display_labels:
            return repr(self)

        display_elements = []
        for item in self._schema.display_labels:
            item_elements = item.split("__")
            if len(item_elements) != 2:
                raise ValidationError("Display Label can only have one level")

            if item_elements[0] not in self._schema.attribute_names:
                raise ValidationError("Only Attribute can be used in Display Label")

            attr = getattr(self, item_elements[0])
            attr_value = getattr(attr, item_elements[1])
            if isinstance(attr_value, Enum):
                display_elements.append(attr_value.value)
            else:
                display_elements.append(attr_value)

        if not display_elements or all(de is None for de in display_elements):
            return ""
        display_label = " ".join([str(de) for de in display_elements])
        if not display_label.strip():
            return repr(self)
        return display_label.strip()

    async def set_human_friendly_id(self, value: list[str] | None) -> None:
        """Set the human friendly ID of this node if one is set. `save()` must be called to commit the change in the database."""
        if self._human_friendly_id is None:
            return

        self._human_friendly_id.set_value(value=value, manually_assigned=True)

    async def set_display_label(self, value: str | None) -> None:
        """Set the display label of this node if one is set. `save()` must be called to commit the change in the database."""
        if self._display_label is None:
            return

        self._display_label.set_value(value=value, manually_assigned=True)

    def _get_parent_relationship_name(self) -> str | None:
        """Return the name of the parent relationship is one is present"""
        for relationship in self._schema.relationships:
            if relationship.kind == RelationshipKind.PARENT:
                return relationship.name

        return None

    async def get_object_template(self, db: InfrahubDatabase) -> CoreObjectTemplate | None:
        object_template: RelationshipManager | None = getattr(self, OBJECT_TEMPLATE_RELATIONSHIP_NAME, None)
        return (
            await object_template.get_peer(db=db, peer_type=CoreObjectTemplate) if object_template is not None else None
        )

    def get_relationships(
        self, kind: RelationshipKind, exclude: Sequence[str] | None = None
    ) -> list[RelationshipSchema]:
        """Return relationships of a given kind with the possiblity to exclude some of them by name."""
        if exclude is None:
            exclude = []

        return [
            relationship
            for relationship in self.get_schema().relationships
            if relationship.name not in exclude and relationship.kind == kind
        ]

    def validate_relationships(self) -> None:
        for name in self._relationships:
            relm: RelationshipManager = getattr(self, name)
            relm.validate()

    async def get_parent_relationship_peer(self, db: InfrahubDatabase, name: str) -> Node | None:
        """When a node has a parent relationship of a given name, this method returns the peer of that relationship."""
        relationship = self.get_schema().get_relationship(name=name)
        if relationship.kind != RelationshipKind.PARENT:
            raise ValueError(f"Relationship '{name}' is not of kind 'parent'")

        relm: RelationshipManager = getattr(self, name)
        return await relm.get_peer(db=db)



from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import Any

import yaml
from pydantic import BaseModel, Field
from typing_extensions import Self
from yaml.parser import ParserError

from .exceptions import FileNotValidError
from .utils import read_file


class InfrahubFileApiVersion(str, Enum):
    V1 = "infrahub.app/v1"


class InfrahubFileKind(str, Enum):
    MENU = "Menu"
    OBJECT = "Object"


class InfrahubFileData(BaseModel):
    api_version: InfrahubFileApiVersion = Field(InfrahubFileApiVersion.V1, alias="apiVersion")
    kind: InfrahubFileKind
    spec: dict
    metadata: dict = Field(default_factory=dict)


class LocalFile(BaseModel):
    identifier: str | None = None
    location: Path
    multiple_documents: bool = False
    document_position: int | None = None
    content: dict | None = None
    valid: bool = True
    error_message: str | None = None


class YamlFile(LocalFile):
    def load_content(self) -> None:
        try:
            self.content = yaml.safe_load(read_file(self.location))
        except FileNotValidError as exc:
            self.error_message = exc.message
            self.valid = False
            return

        except yaml.YAMLError:
            self.error_message = "Invalid YAML/JSON file"
            self.valid = False
            return

        if not self.content:
            self.error_message = "Empty YAML/JSON file"
            self.valid = False

    def validate_content(self) -> None:
        pass

    @classmethod
    def init(cls, location: Path, multiple_documents: bool, content: dict | None) -> Self:
        if not content:
            return cls._file_is_empty(path=location, has_multiple_document=multiple_documents)

        return cls(location=location, multiple_documents=multiple_documents, content=content)

    @classmethod
    def _file_is_empty(cls, path: Path, has_multiple_document: bool) -> Self:
        return cls(
            location=path, multiple_documents=has_multiple_document, error_message="Invalid YAML/JSON file", valid=False
        )

    @classmethod
    def _file_is_invalid(cls, path: Path, has_multiple_document: bool) -> Self:
        return cls(
            location=path,
            multiple_documents=has_multiple_document,
            error_message="Invalid YAML/JSON file",
            valid=False,
        )

    @classmethod
    def load_file_from_disk(cls, path: Path) -> list[Self]:
        yaml_files: list[Self] = []

        try:
            file_content = read_file(path)

            has_multiple_document = bool(file_content.count("---") > 1)

            if has_multiple_document:
                yaml_files.extend(
                    cls.init(location=path, multiple_documents=has_multiple_document, content=content)
                    for content in yaml.safe_load_all(file_content)
                )

            else:
                yaml_files.append(
                    cls.init(
                        location=path, multiple_documents=has_multiple_document, content=yaml.safe_load(file_content)
                    )
                )

        except FileNotValidError as exc:
            yaml_files.append(
                cls(location=path, multiple_documents=has_multiple_document, error_message=exc.message, valid=False)
            )
        except (yaml.YAMLError, ParserError):
            yaml_files.append(cls._file_is_invalid(path, has_multiple_document))

        if has_multiple_document:
            for idx, file in enumerate(yaml_files):
                file.document_position = idx + 1

        return yaml_files

    @classmethod
    def load_from_disk(cls, paths: list[Path]) -> list[Self]:
        yaml_files: list[Self] = []
        file_extensions = {".yaml", ".yml", ".json"}  # FIXME: .json is not a YAML file, should be removed

        for file_path in paths:
            if not file_path.exists():
                # Check if the provided path exists, relevant for the first call coming from the user
                raise FileNotValidError(name=str(file_path), message=f"{file_path} does not exist!")
            if file_path.is_file():
                if file_path.suffix in file_extensions:
                    yaml_files.extend(cls.load_file_from_disk(path=file_path))
                # else: silently skip files with unrelevant extensions (e.g. .md, .py...)
            elif file_path.is_dir():
                # Introduce recursion to handle sub-folders
                sub_paths = [Path(sub_file_path) for sub_file_path in file_path.glob("*")]
                sub_paths = sorted(sub_paths, key=lambda p: p.name)
                yaml_files.extend(cls.load_from_disk(paths=sub_paths))
            # else: skip non-file, non-dir (e.g., symlink...)

        return yaml_files


class InfrahubFile(YamlFile):
    _data: InfrahubFileData | None = None

    @property
    def data(self) -> InfrahubFileData:
        if not self._data:
            raise ValueError("_data hasn't been initialized yet")
        return self._data

    @property
    def version(self) -> InfrahubFileApiVersion:
        return self.data.api_version

    @property
    def kind(self) -> InfrahubFileKind:
        return self.data.kind

    def validate_content(self) -> None:
        if not self.content:
            raise ValueError("Content hasn't been loaded yet")
        self._data = InfrahubFileData(**self.content)


class SchemaFile(YamlFile):
    @property
    def payload(self) -> dict[str, Any]:
        return self.content or {}



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Field, Int, List, NonNull, ObjectType, String

from infrahub.core.manager import NodeManager
from infrahub.core.protocols import InternalAccountToken
from infrahub.exceptions import PermissionDeniedError
from infrahub.graphql.field_extractor import extract_graphql_fields

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.graphql.initialization import GraphqlContext


class AccountTokenNode(ObjectType):
    id = Field(String, required=True)
    name = Field(String, required=False)
    expiration = Field(String, required=False)


class AccountTokenEdge(ObjectType):
    node = Field(AccountTokenNode, required=True)


class AccountTokenEdges(ObjectType):
    count = Field(Int, required=True)
    edges = Field(List(of_type=NonNull(AccountTokenEdge), required=True), required=True)


async def resolve_account_tokens(
    root: dict,  # noqa: ARG001
    info: GraphQLResolveInfo,
    limit: int = 10,
    offset: int = 0,
) -> dict:
    graphql_context: GraphqlContext = info.context

    if not graphql_context.account_session:
        raise ValueError("An account_session is mandatory to execute this query")

    if not graphql_context.account_session.authenticated_by_jwt:
        raise PermissionDeniedError("This operation requires authentication with a JWT token")

    fields = extract_graphql_fields(info)

    filters = {"account__ids": [graphql_context.account_session.account_id]}
    response: dict[str, Any] = {}
    if "count" in fields:
        response["count"] = await NodeManager.count(db=graphql_context.db, schema=InternalAccountToken, filters=filters)
    if "edges" in fields:
        objs = await NodeManager.query(
            db=graphql_context.db, schema=InternalAccountToken, filters=filters, limit=limit, offset=offset
        )
        response["edges"] = [
            {"node": {"id": obj.id, "name": obj.name.value, "expiration": obj.expiration.value}} for obj in objs
        ]

    return response


AccountToken = Field(
    AccountTokenEdges,
    limit=Int(required=False),
    offset=Int(required=False),
    resolver=resolve_account_tokens,
    required=True,
)


class AccountGlobalPermissionNode(ObjectType):
    id = Field(String, required=True)
    display_label = Field(String, required=True)
    description = Field(String, required=False)
    name = Field(String, required=True)
    action = Field(String, required=True)
    decision = Field(String, required=True)
    identifier = Field(String, required=True)


class AccountObjectPermissionNode(ObjectType):
    id = Field(String, required=True)
    display_label = Field(String, required=True)
    description = Field(String, required=False)
    namespace = Field(String, required=True)
    name = Field(String, required=True)
    action = Field(String, required=True)
    decision = Field(String, required=True)
    identifier = Field(String, required=True)


class AccountGlobalPermissionEdge(ObjectType):
    node = Field(AccountGlobalPermissionNode, required=True)


class AccountObjectPermissionEdge(ObjectType):
    node = Field(AccountObjectPermissionNode, required=True)


class AccountGlobalPermissionEdges(ObjectType):
    count = Field(Int, required=True)
    edges = Field(List(of_type=NonNull(AccountGlobalPermissionEdge), required=True), required=True)


class AccountObjectPermissionEdges(ObjectType):
    count = Field(Int, required=True)
    edges = Field(List(of_type=NonNull(AccountObjectPermissionEdge), required=True), required=True)


class AccountPermissionsEdges(ObjectType):
    global_permissions = Field(AccountGlobalPermissionEdges, required=False)
    object_permissions = Field(AccountObjectPermissionEdges, required=False)


async def resolve_account_permissions(
    root: dict,  # noqa: ARG001
    info: GraphQLResolveInfo,
) -> dict:
    graphql_context: GraphqlContext = info.context

    if not graphql_context.account_session:
        raise ValueError("An account_session is mandatory to execute this query")

    fields = extract_graphql_fields(info)

    response: dict[str, dict[str, Any]] = {}
    if "global_permissions" in fields:
        global_list = graphql_context.active_permissions.permissions["global_permissions"]
        response["global_permissions"] = {"count": len(global_list)}
        response["global_permissions"]["edges"] = [
            {
                "node": {
                    "id": obj.id,
                    "display_label": str(obj),
                    "description": obj.description,
                    "action": obj.action,
                    "decision": obj.decision,
                    "identifier": str(obj),
                }
            }
            for obj in global_list
        ]
    if "object_permissions" in fields:
        object_list = graphql_context.active_permissions.permissions["object_permissions"]
        response["object_permissions"] = {"count": len(object_list)}
        response["object_permissions"]["edges"] = [
            {
                "node": {
                    "id": obj.id,
                    "display_label": str(obj),
                    "description": obj.description,
                    "namespace": obj.namespace,
                    "name": obj.name,
                    "action": obj.action,
                    "decision": obj.decision,
                    "identifier": str(obj),
                }
            }
            for obj in object_list
        ]
    return response


AccountPermissions = Field(AccountPermissionsEdges, resolver=resolve_account_permissions, required=True)



from __future__ import annotations

import asyncio
import difflib
from pathlib import Path
from typing import TYPE_CHECKING, Any

import jinja2
import ujson
from httpx import HTTPStatusError

from ...template import Jinja2Template
from ...template.exceptions import JinjaTemplateError
from ..exceptions import OutputMatchError
from ..models import InfrahubInputOutputTest, InfrahubTestExpectedResult
from .base import InfrahubItem

if TYPE_CHECKING:
    import pytest


class InfrahubJinja2Item(InfrahubItem):
    def _get_jinja2(self) -> Jinja2Template:
        return Jinja2Template(
            template=Path(self.resource_config.template_path),  # type: ignore[attr-defined]
            template_directory=Path(self.session.infrahub_config_path.parent),  # type: ignore[attr-defined]
        )

    def get_jinja2_environment(self) -> jinja2.Environment:
        jinja2_template = self._get_jinja2()
        return jinja2_template.get_environment()

    def get_jinja2_template(self) -> jinja2.Template:
        jinja2_template = self._get_jinja2()
        return jinja2_template.get_template()

    def render_jinja2_template(self, variables: dict[str, Any]) -> str | None:
        jinja2_template = self._get_jinja2()

        try:
            return asyncio.run(jinja2_template.render(variables=variables))
        except JinjaTemplateError as exc:
            if self.test.expect == InfrahubTestExpectedResult.PASS:
                raise exc
            return None

    def get_result_differences(self, computed: Any) -> str | None:
        if not isinstance(self.test.spec, InfrahubInputOutputTest) or not self.test.spec.output or computed is None:
            return None

        differences = difflib.unified_diff(
            self.test.spec.get_output_data().splitlines(),
            computed.splitlines(),
            fromfile="expected",
            tofile="rendered",
            lineterm="",
        )
        return "\n".join(differences)

    def repr_failure(self, excinfo: pytest.ExceptionInfo, style: str | None = None) -> str:
        if isinstance(excinfo.value, HTTPStatusError):
            try:
                response_content = ujson.dumps(excinfo.value.response.json(), indent=4, sort_keys=True)
            except ujson.JSONDecodeError:
                response_content = excinfo.value.response.text
            return "\n".join(
                [
                    f"Failed {excinfo.value.request.method} on {excinfo.value.request.url}",
                    f"Status code: {excinfo.value.response.status_code}",
                    f"Response: {response_content}",
                ]
            )

        if isinstance(excinfo.value, jinja2.TemplateSyntaxError):
            return "\n".join(["Syntax error detected in the template", excinfo.value.message or ""])

        if isinstance(excinfo.value, OutputMatchError):
            return f"{excinfo.value.message}\n{excinfo.value.differences}"

        return super().repr_failure(excinfo, style=style)


class InfrahubJinja2TransformSmokeItem(InfrahubJinja2Item):
    def runtest(self) -> None:
        file_path: Path = self.session.infrahub_config_path.parent / self.resource_config.template_path  # type: ignore[attr-defined]
        self.get_jinja2_environment().parse(file_path.read_text(encoding="utf-8"), filename=file_path.name)


class InfrahubJinja2TransformUnitRenderItem(InfrahubJinja2Item):
    def runtest(self) -> None:
        computed = self.render_jinja2_template(self.test.spec.get_input_data())  # type: ignore[union-attr]
        differences = self.get_result_differences(computed)

        if computed is not None and differences and self.test.expect == InfrahubTestExpectedResult.PASS:
            raise OutputMatchError(name=self.name, differences=differences)

    def repr_failure(self, excinfo: pytest.ExceptionInfo, style: str | None = None) -> str:
        if isinstance(excinfo.value, (JinjaTemplateError)):
            return str(excinfo.value.message)

        return super().repr_failure(excinfo, style=style)


class InfrahubJinja2TransformIntegrationItem(InfrahubJinja2Item):
    def runtest(self) -> None:
        graphql_result = self.session.infrahub_client.query_gql_query(  # type: ignore[attr-defined]
            self.resource_config.query,  # type: ignore[attr-defined]
            variables=self.test.spec.get_variables_data(),  # type: ignore[union-attr]
        )
        computed = self.render_jinja2_template(graphql_result)
        differences = self.get_result_differences(computed)

        if computed is not None and differences and self.test.expect == InfrahubTestExpectedResult.PASS:
            raise OutputMatchError(name=self.name, differences=differences)



from __future__ import annotations

from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO, cast, overload

import anyio
import httpx

from .exceptions import AuthenticationError, NodeNotFoundError, ServerNotReachableError

if TYPE_CHECKING:
    from .client import InfrahubClient, InfrahubClientSync


@dataclass
class PreparedFile:
    file_object: BinaryIO | None
    filename: str | None
    should_close: bool


class FileHandlerBase:
    """Base class for file handling operations.

    Provides common functionality for both async and sync file handlers, including upload preparation and error handling.
    """

    @staticmethod
    async def prepare_upload(content: bytes | Path | BinaryIO | None, name: str | None = None) -> PreparedFile:
        """Prepare file content for upload (async version).

        Converts various content types to a consistent BinaryIO interface for streaming uploads.
        For Path inputs, opens the file handle in a thread pool to avoid blocking the event loop.
        The actual file reading is streamed by httpx during the HTTP request.

        Args:
            content: The file content as bytes, a Path to a file, or a file-like object.
                     Can be None if no file is set.
            name: Optional filename. If not provided and content is a Path,
                  the filename will be derived from the path.

        Returns:
            A PreparedFile containing the file object, filename, and whether it should be closed.
        """
        if content is None:
            return PreparedFile(file_object=None, filename=None, should_close=False)

        if name is None and isinstance(content, Path):
            name = content.name

        filename = name or "uploaded_file"

        if isinstance(content, bytes):
            return PreparedFile(file_object=BytesIO(content), filename=filename, should_close=False)
        if isinstance(content, Path):
            # Open file in thread pool to avoid blocking the event loop
            # Returns a sync file handle that httpx can stream from in chunks
            file_obj = await anyio.to_thread.run_sync(content.open, "rb")
            return PreparedFile(file_object=cast("BinaryIO", file_obj), filename=filename, should_close=True)

        # At this point, content must be a BinaryIO (file-like object)
        return PreparedFile(file_object=cast("BinaryIO", content), filename=filename, should_close=False)

    @staticmethod
    def prepare_upload_sync(content: bytes | Path | BinaryIO | None, name: str | None = None) -> PreparedFile:
        """Prepare file content for upload (sync version).

        Converts various content types to a consistent BinaryIO interface for streaming uploads.

        Args:
            content: The file content as bytes, a Path to a file, or a file-like object.
                     Can be None if no file is set.
            name: Optional filename. If not provided and content is a Path,
                  the filename will be derived from the path.

        Returns:
            A PreparedFile containing the file object, filename, and whether it should be closed.
        """
        if content is None:
            return PreparedFile(file_object=None, filename=None, should_close=False)

        if name is None and isinstance(content, Path):
            name = content.name

        filename = name or "uploaded_file"

        if isinstance(content, bytes):
            return PreparedFile(file_object=BytesIO(content), filename=filename, should_close=False)
        if isinstance(content, Path):
            return PreparedFile(file_object=content.open("rb"), filename=filename, should_close=True)

        # At this point, content must be a BinaryIO (file-like object)
        return PreparedFile(file_object=cast("BinaryIO", content), filename=filename, should_close=False)

    @staticmethod
    def handle_error_response(exc: httpx.HTTPStatusError) -> None:
        """Handle HTTP error responses for file operations.

        Args:
            exc: The HTTP status error from httpx.

        Raises:
            AuthenticationError: If authentication fails (401/403).
            NodeNotFoundError: If the file/node is not found (404).
            httpx.HTTPStatusError: For other HTTP errors.
        """
        if exc.response.status_code in {401, 403}:
            response = exc.response.json()
            errors = response.get("errors", [])
            messages = [error.get("message") for error in errors]
            raise AuthenticationError(" | ".join(messages)) from exc
        if exc.response.status_code == 404:
            response = exc.response.json()
            detail = response.get("detail", "File not found")
            raise NodeNotFoundError(node_type="FileObject", identifier=detail) from exc
        raise exc

    @staticmethod
    def handle_response(resp: httpx.Response) -> bytes:
        """Handle the HTTP response and return file content as bytes.

        Args:
            resp: The HTTP response from httpx.

        Returns:
            The file content as bytes.

        Raises:
            AuthenticationError: If authentication fails.
            NodeNotFoundError: If the file is not found.
        """
        try:
            resp.raise_for_status()
        except httpx.HTTPStatusError as exc:
            FileHandlerBase.handle_error_response(exc=exc)
        return resp.content


class FileHandler(FileHandlerBase):
    """Async file handler for download operations.

    Handles file downloads with support for streaming to disk
    for memory-efficient handling of large files.
    """

    def __init__(self, client: InfrahubClient) -> None:
        """Initialize the async file handler.

        Args:
            client: The async Infrahub client instance.
        """
        self._client = client

    def _build_url(self, node_id: str, branch: str | None) -> str:
        """Build the download URL for a file.

        Args:
            node_id: The ID of the FileObject node.
            branch: Optional branch name.

        Returns:
            The complete URL for downloading the file.
        """
        url = f"{self._client.address}/api/storage/files/{node_id}"
        if branch:
            url = f"{url}?branch={branch}"
        return url

    @overload
    async def download(self, node_id: str, branch: str | None) -> bytes: ...

    @overload
    async def download(self, node_id: str, branch: str | None, dest: Path) -> int: ...

    @overload
    async def download(self, node_id: str, branch: str | None, dest: None) -> bytes: ...

    async def download(self, node_id: str, branch: str | None, dest: Path | None = None) -> bytes | int:
        """Download file content from a FileObject node.

        Args:
            node_id: The ID of the FileObject node.
            branch: Optional branch name. Uses client default if not provided.
            dest: Optional destination path. If provided, streams to disk.

        Returns:
            If dest is None: The file content as bytes.
            If dest is provided: The number of bytes written.

        Raises:
            ServerNotReachableError: If the server is not reachable.
            AuthenticationError: If authentication fails.
            NodeNotFoundError: If the node/file is not found.
        """
        effective_branch = branch or self._client.default_branch
        url = self._build_url(node_id=node_id, branch=effective_branch)

        if dest is not None:
            return await self._stream_to_file(url=url, dest=dest)

        try:
            resp = await self._client._get(url=url)
        except ServerNotReachableError:
            self._client.log.error(f"Unable to connect to {self._client.address}")
            raise

        return self.handle_response(resp=resp)

    async def _stream_to_file(self, url: str, dest: Path) -> int:
        """Stream download directly to a file without loading into memory.

        Args:
            url: The URL to download from.
            dest: The destination path to write to.

        Returns:
            The number of bytes written to the file.

        Raises:
            ServerNotReachableError: If the server is not reachable.
            AuthenticationError: If authentication fails.
            NodeNotFoundError: If the file is not found.
        """
        try:
            async with self._client._get_streaming(url=url) as resp:
                try:
                    resp.raise_for_status()
                except httpx.HTTPStatusError as exc:
                    # Need to read the response body for error details
                    await resp.aread()
                    self.handle_error_response(exc=exc)

                bytes_written = 0
                async with await anyio.Path(dest).open("wb") as f:
                    async for chunk in resp.aiter_bytes(chunk_size=65536):
                        await f.write(chunk)
                        bytes_written += len(chunk)
                return bytes_written
        except ServerNotReachableError:
            self._client.log.error(f"Unable to connect to {self._client.address}")
            raise


class FileHandlerSync(FileHandlerBase):
    """Sync file handler for download operations.

    Handles file downloads with support for streaming to disk
    for memory-efficient handling of large files.
    """

    def __init__(self, client: InfrahubClientSync) -> None:
        """Initialize the sync file handler.

        Args:
            client: The sync Infrahub client instance.
        """
        self._client = client

    def _build_url(self, node_id: str, branch: str | None) -> str:
        """Build the download URL for a file.

        Args:
            node_id: The ID of the FileObject node.
            branch: Optional branch name.

        Returns:
            The complete URL for downloading the file.
        """
        url = f"{self._client.address}/api/storage/files/{node_id}"
        if branch:
            url = f"{url}?branch={branch}"
        return url

    @overload
    def download(self, node_id: str, branch: str | None) -> bytes: ...

    @overload
    def download(self, node_id: str, branch: str | None, dest: Path) -> int: ...

    @overload
    def download(self, node_id: str, branch: str | None, dest: None) -> bytes: ...

    def download(self, node_id: str, branch: str | None, dest: Path | None = None) -> bytes | int:
        """Download file content from a FileObject node.

        Args:
            node_id: The ID of the FileObject node.
            branch: Optional branch name. Uses client default if not provided.
            dest: Optional destination path. If provided, streams to disk.

        Returns:
            If dest is None: The file content as bytes.
            If dest is provided: The number of bytes written.

        Raises:
            ServerNotReachableError: If the server is not reachable.
            AuthenticationError: If authentication fails.
            NodeNotFoundError: If the node/file is not found.
        """
        effective_branch = branch or self._client.default_branch
        url = self._build_url(node_id=node_id, branch=effective_branch)

        if dest is not None:
            return self._stream_to_file(url=url, dest=dest)

        try:
            resp = self._client._get(url=url)
        except ServerNotReachableError:
            self._client.log.error(f"Unable to connect to {self._client.address}")
            raise

        return self.handle_response(resp=resp)

    def _stream_to_file(self, url: str, dest: Path) -> int:
        """Stream download directly to a file without loading into memory.

        Args:
            url: The URL to download from.
            dest: The destination path to write to.

        Returns:
            The number of bytes written to the file.

        Raises:
            ServerNotReachableError: If the server is not reachable.
            AuthenticationError: If authentication fails.
            NodeNotFoundError: If the file is not found.
        """
        try:
            with self._client._get_streaming(url=url) as resp:
                try:
                    resp.raise_for_status()
                except httpx.HTTPStatusError as exc:
                    # Need to read the response body for error details
                    resp.read()
                    self.handle_error_response(exc=exc)

                bytes_written = 0
                with dest.open("wb") as f:
                    for chunk in resp.iter_bytes(chunk_size=65536):
                        f.write(chunk)
                        bytes_written += len(chunk)
                return bytes_written
        except ServerNotReachableError:
            self._client.log.error(f"Unable to connect to {self._client.address}")
            raise



from abc import abstractmethod

from infrahub.core.timestamp import Timestamp


class MetadataInterface:
    @abstractmethod
    def _set_created_at(self, value: Timestamp | None) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _set_created_by(self, value: str | None) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _set_updated_at(self, value: Timestamp | None) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _set_updated_by(self, value: str | None) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _get_created_at(self) -> Timestamp | None:
        raise NotImplementedError()

    @abstractmethod
    def _get_created_by(self) -> str | None:
        raise NotImplementedError()

    @abstractmethod
    def _get_updated_at(self) -> Timestamp | None:
        raise NotImplementedError()

    @abstractmethod
    def _get_updated_by(self) -> str | None:
        raise NotImplementedError()



from __future__ import annotations

from datetime import timedelta
from enum import StrEnum
from typing import TYPE_CHECKING, Any, TypeVar

from prefect.client.schemas.objects import StateType  # noqa: TC002
from prefect.events.actions import ChangeFlowRunState, RunDeployment
from prefect.events.schemas.automations import Automation, Posture
from prefect.events.schemas.automations import EventTrigger as PrefectEventTrigger
from prefect.events.schemas.events import ResourceSpecification
from pydantic import BaseModel, Field

from infrahub import __version__
from infrahub.workflows.models import WorkflowDefinition  # noqa: TC001

from .constants import NAME_SEPARATOR

if TYPE_CHECKING:
    from uuid import UUID

T = TypeVar("T", bound="TriggerDefinition")


class TriggerComparison(StrEnum):
    MATCH = "match"  # Expected trigger and actual trigger is identical
    REFRESH = "refresh"  # The branch parameters doesn't match, the hash does, refresh in Prefect but don't run triggers
    UPDATE = "update"  # Neither branch or other data points match, update in Prefect and run triggers

    @property
    def update_prefect(self) -> bool:
        return self in {TriggerComparison.REFRESH, TriggerComparison.UPDATE}


class TriggerSetupReport(BaseModel):
    created: list[TriggerDefinition] = Field(default_factory=list)
    refreshed: list[TriggerDefinition] = Field(default_factory=list)
    updated: list[TriggerDefinition] = Field(default_factory=list)
    deleted: list[Automation] = Field(default_factory=list)
    unchanged: list[TriggerDefinition] = Field(default_factory=list)

    @property
    def in_use_count(self) -> int:
        return len(self.created + self.updated + self.unchanged + self.refreshed)

    def add_with_comparison(self, trigger: TriggerDefinition, comparison: TriggerComparison) -> None:
        match comparison:
            case TriggerComparison.UPDATE:
                self.updated.append(trigger)
            case TriggerComparison.REFRESH:
                self.refreshed.append(trigger)
            case TriggerComparison.MATCH:
                self.unchanged.append(trigger)

    def _created_triggers_with_type(self, trigger_type: type[T]) -> list[T]:
        return [trigger for trigger in self.created if isinstance(trigger, trigger_type)]

    def _refreshed_triggers_with_type(self, trigger_type: type[T]) -> list[T]:
        return [trigger for trigger in self.refreshed if isinstance(trigger, trigger_type)]

    def _unchanged_triggers_with_type(self, trigger_type: type[T]) -> list[T]:
        return [trigger for trigger in self.unchanged if isinstance(trigger, trigger_type)]

    def _updated_triggers_with_type(self, trigger_type: type[T]) -> list[T]:
        return [trigger for trigger in self.updated if isinstance(trigger, trigger_type)]

    def triggers_with_type(self, trigger_type: type[T]) -> list[T]:
        """Return all triggers that match the specified type.

        Args:
            trigger_type: A TriggerDefinition class or subclass to filter by

        Returns:
            List of triggers of the specified type from all categories
        """
        created = self._created_triggers_with_type(trigger_type=trigger_type)
        updated = self._updated_triggers_with_type(trigger_type=trigger_type)
        refreshed = self._refreshed_triggers_with_type(trigger_type=trigger_type)
        unchanged = self._unchanged_triggers_with_type(trigger_type=trigger_type)
        return created + updated + refreshed + unchanged

    def modified_triggers_with_type(self, trigger_type: type[T]) -> list[T]:
        """Return all created and updated triggers that match the specified type.

        Args:
            trigger_type: A TriggerDefinition class or subclass to filter by

        Returns:
            List of triggers of the specified type from both created and updated lists
        """
        created = self._created_triggers_with_type(trigger_type=trigger_type)
        updated = self._updated_triggers_with_type(trigger_type=trigger_type)
        return created + updated


class TriggerType(StrEnum):
    ACTION_TRIGGER_RULE = "action_trigger_rule"
    BUILTIN = "builtin"
    WEBHOOK = "webhook"
    COMPUTED_ATTR_JINJA2 = "computed_attr_jinja2"
    COMPUTED_ATTR_PYTHON = "computed_attr_python"
    COMPUTED_ATTR_PYTHON_QUERY = "computed_attr_python_query"
    DISPLAY_LABEL_JINJA2 = "display_label_jinja2"
    HUMAN_FRIENDLY_ID = "human_friendly_id"
    PROFILE = "profile"
    # OBJECT = "object"

    @property
    def is_branch_specific(self) -> bool:
        return self in {
            TriggerType.COMPUTED_ATTR_JINJA2,
            TriggerType.COMPUTED_ATTR_PYTHON,
            TriggerType.COMPUTED_ATTR_PYTHON_QUERY,
            TriggerType.DISPLAY_LABEL_JINJA2,
            TriggerType.HUMAN_FRIENDLY_ID,
        }


def _match_related_dict() -> dict:
    # Make Mypy happy as match related is a dict[str, Any] | list[dict[str, Any]]
    return {}


class EventTrigger(BaseModel):
    events: set = Field(default_factory=set)
    match: dict[str, Any] = Field(default_factory=dict)
    match_related: dict[str, Any] | list[dict[str, Any]] = Field(default_factory=_match_related_dict)

    def get_prefect(self) -> PrefectEventTrigger:
        return PrefectEventTrigger(
            posture=Posture.Reactive,
            expect=self.events,
            within=timedelta(0),
            match=ResourceSpecification(self.match),
            match_related=self.related_resource_specification,
            threshold=1,
        )

    @property
    def related_resource_specification(self) -> ResourceSpecification | list[ResourceSpecification]:
        if isinstance(self.match_related, dict):
            return ResourceSpecification(self.match_related)

        if len(self.match_related) == 1:
            return ResourceSpecification(self.match_related[0])

        return [ResourceSpecification(related_match) for related_match in self.match_related]


class ProactiveEventTrigger(EventTrigger):
    """A proactive event trigger that fires when expected events do NOT occur within a time window.

    Unlike EventTrigger which uses Reactive posture (fires when events occur),
    ProactiveEventTrigger uses Proactive posture to detect missing events.
    """

    after: set[str] = Field(default_factory=set)
    for_each: set[str] = Field(default_factory=set)
    threshold: int = 1
    within: timedelta = Field(default_factory=lambda: timedelta(seconds=90))

    def get_prefect(self) -> PrefectEventTrigger:
        return PrefectEventTrigger(
            posture=Posture.Proactive,
            after=self.after,
            expect=self.events,
            match=ResourceSpecification(self.match),
            for_each=self.for_each,
            threshold=self.threshold,
            within=self.within,
        )


class ChangeFlowRunStateAction(BaseModel):
    """Action to change the state of a flow run.

    Used for system automations that need to modify flow run states,
    such as crashing zombie flows that have stopped sending heartbeats.
    """

    state: StateType
    message: str = ""

    def get_prefect(self, _mapping: dict[str, UUID] | None = None) -> ChangeFlowRunState:
        """Get the Prefect ChangeFlowRunState action.

        Args:
            _mapping: Not used for this action type, but included for interface compatibility.

        Returns:
            A Prefect ChangeFlowRunState action.
        """
        return ChangeFlowRunState(  # type: ignore[call-arg]
            state=self.state,
            message=self.message,
        )


class ExecuteWorkflow(BaseModel):
    workflow: WorkflowDefinition
    parameters: dict[str, Any] = Field(default_factory=dict)

    @property
    def name(self) -> str:
        return self.workflow.name

    def get_prefect(self, mapping: dict[str, UUID]) -> RunDeployment:
        deployment_id = mapping[self.name]
        return self.get(deployment_id)

    def get(self, id: UUID) -> RunDeployment:
        return RunDeployment(
            source="selected",
            deployment_id=id,
            parameters=self.parameters,
            job_variables={},
        )

    def validate_parameters(self) -> None:
        if not self.parameters:
            return

        workflow_params = self.workflow.get_parameters()
        workflow_required_params = [p.name for p in workflow_params.values() if p.required]
        trigger_params = list(self.parameters.keys())

        missing_required_params = set(workflow_required_params) - set(trigger_params)
        wrong_params = set(trigger_params) - set(workflow_params)

        if missing_required_params:
            raise ValueError(
                f"Missing required parameters: {missing_required_params} for workflow {self.workflow.name}"
            )

        if wrong_params:
            raise ValueError(f"Workflow {self.workflow.name} doesn't support parameters: {wrong_params}")


# Type alias for all trigger action types
TriggerActionType = ExecuteWorkflow | ChangeFlowRunStateAction


class TriggerDefinition(BaseModel):
    name: str
    type: TriggerType
    previous_names: set = Field(default_factory=set)
    description: str = ""
    trigger: EventTrigger
    actions: list[TriggerActionType]

    def get_deployment_names(self) -> list[str]:
        """Return the name of all deployments used by this trigger"""
        return [action.name for action in self.actions if isinstance(action, ExecuteWorkflow)]

    def get_description(self) -> str:
        return f"Automation for Trigger {self.name} of type {self.type.value} (v{__version__})"

    def generate_name(self) -> str:
        return f"{self.type.value}{NAME_SEPARATOR}{self.name}"

    def validate_actions(self) -> None:
        for action in self.actions:
            if isinstance(action, ExecuteWorkflow):
                action.validate_parameters()


class TriggerBranchDefinition(TriggerDefinition):
    branch: str

    def generate_name(self) -> str:
        return f"{self.type.value}{NAME_SEPARATOR}{self.branch}{NAME_SEPARATOR}{self.name}"


class BuiltinTriggerDefinition(TriggerDefinition):
    type: TriggerType = TriggerType.BUILTIN


class SystemTriggerDefinition(BuiltinTriggerDefinition):
    """A trigger definition for system-level Prefect automations.

    Unlike other TriggerDefinitions which execute Infrahub workflows, SystemTriggerDefinition
    is designed for Prefect system automations that don't require workflow deployments,
    such as crashing zombie flows.

    Uses ChangeFlowRunStateAction for actions (not ExecuteWorkflow).
    """

    def get_description(self) -> str:
        return f"System Automation for {self.name} (v{__version__})"



from __future__ import annotations

from graphene import DateTime, Field, List, NonNull, ObjectType, String


class RelationshipPeer(ObjectType):
    id = String(required=False)
    kind = String(required=False)


class InfrahubRelationshipMetaObject(ObjectType):
    updated_by = String(required=False, description="User that last modified the relationship")
    updated_at = DateTime(
        required=False,
        description="Date/Time when the relationship was last modified by a user or a system task",
    )


class Relationship(InfrahubRelationshipMetaObject):
    id = String(required=False)
    identifier = String(required=False)
    peers = List(NonNull(RelationshipPeer))


class RelationshipNode(ObjectType):
    node = Field(Relationship, required=True)



from typing import TYPE_CHECKING, Awaitable, Callable, Sequence

from prefect import get_run_logger, task
from prefect.automations import AutomationCore
from prefect.cache_policies import NONE
from prefect.client.orchestration import PrefectClient, get_client
from prefect.client.schemas.filters import DeploymentFilter, DeploymentFilterName
from prefect.events.schemas.automations import Automation
from prefect.exceptions import PrefectHTTPStatusError

from infrahub import lock
from infrahub.database import InfrahubDatabase
from infrahub.trigger.models import TriggerDefinition

from .models import TriggerComparison, TriggerSetupReport, TriggerType

if TYPE_CHECKING:
    from uuid import UUID


def compare_automations(
    target: AutomationCore, existing: Automation, trigger_type: TriggerType | None, force_update: bool = False
) -> TriggerComparison:
    """Compare an AutomationCore with an existing Automation object to identify if they are identical,
    if it's a branch specific automation and the branch filter may be different, or if they are different.
    """

    if force_update:
        return TriggerComparison.UPDATE

    target_dump = target.model_dump(exclude_defaults=True, exclude_none=True)
    existing_dump = existing.model_dump(exclude_defaults=True, exclude_none=True, exclude={"id"})

    if target_dump == existing_dump:
        return TriggerComparison.MATCH

    if not trigger_type or not trigger_type.is_branch_specific:
        return TriggerComparison.UPDATE

    if target.description == existing.description:
        # If only the branch related info is different, we consider it a refresh
        return TriggerComparison.REFRESH

    return TriggerComparison.UPDATE


@task(name="trigger-setup-specific", task_run_name="Setup triggers of a specific kind", cache_policy=NONE)  # type: ignore[arg-type]
async def setup_triggers_specific(
    gatherer: Callable[[InfrahubDatabase | None], Awaitable[Sequence[TriggerDefinition]]],
    trigger_type: TriggerType,
    db: InfrahubDatabase | None = None,
) -> TriggerSetupReport:
    async with lock.registry.get(
        name=f"configure-action-rules-{trigger_type.value}", namespace="trigger-rules", local=False
    ):
        if db:
            async with db.start_session(read_only=True) as dbs:
                triggers = await gatherer(dbs)
        else:
            triggers = await gatherer(db)
        async with get_client(sync_client=False) as prefect_client:
            return await setup_triggers(
                client=prefect_client,
                triggers=triggers,
                trigger_type=trigger_type,
            )  # type: ignore[misc]


@task(name="trigger-setup", task_run_name="Setup triggers", cache_policy=NONE)
async def setup_triggers(
    client: PrefectClient,
    triggers: Sequence[TriggerDefinition],
    trigger_type: TriggerType | None = None,
    force_update: bool = False,
) -> TriggerSetupReport:
    log = get_run_logger()

    report = TriggerSetupReport()

    trigger_log_message = f"triggers of type {trigger_type.value}" if trigger_type else "all triggers"
    log.debug(f"Setting up {trigger_log_message}")

    # -------------------------------------------------------------
    # Retrieve existing Deployments and Automation from the server
    # -------------------------------------------------------------
    deployment_names = list({name for trigger in triggers for name in trigger.get_deployment_names()})
    deployments = {
        item.name: item
        for item in await client.read_deployments(
            deployment_filter=DeploymentFilter(name=DeploymentFilterName(any_=deployment_names))
        )
    }
    deployments_mapping: dict[str, UUID] = {name: item.id for name, item in deployments.items()}

    existing_automations = {item.name: item for item in await gather_all_automations(client=client)}
    if trigger_type:
        # If a trigger type is provided, narrow down the list of existing triggers to know which one to delete
        existing_automations = {
            automation_name: automation
            for automation_name, automation in existing_automations.items()
            if automation_name.startswith(f"{trigger_type.value}::")
        }

    trigger_names = [trigger.generate_name() for trigger in triggers]
    automation_names = list(existing_automations.keys())

    log.debug(f"{len(automation_names)} existing triggers ({automation_names})")
    log.debug(f"{len(trigger_names)} triggers to configure ({trigger_names})")

    to_delete = set(automation_names) - set(trigger_names)
    log.debug(f"{len(to_delete)} triggers to delete ({to_delete})")

    # -------------------------------------------------------------
    # Create or Update all triggers
    # -------------------------------------------------------------
    for trigger in triggers:
        automation = AutomationCore(
            name=trigger.generate_name(),
            description=trigger.get_description(),
            enabled=True,
            trigger=trigger.trigger.get_prefect(),
            actions=[action.get_prefect(deployments_mapping) for action in trigger.actions],
        )

        existing_automation = existing_automations.get(trigger.generate_name())

        if existing_automation:
            trigger_comparison = compare_automations(
                target=automation, existing=existing_automation, trigger_type=trigger_type, force_update=force_update
            )
            if trigger_comparison.update_prefect:
                await client.update_automation(automation_id=existing_automation.id, automation=automation)
                log.info(f"{trigger.generate_name()} Updated")
            report.add_with_comparison(trigger, trigger_comparison)
        else:
            await client.create_automation(automation=automation)
            log.info(f"{trigger.generate_name()} Created")
            report.created.append(trigger)

    # -------------------------------------------------------------
    # Delete Triggers that shouldn't be there
    # -------------------------------------------------------------
    for item_to_delete in to_delete:
        existing_automation = existing_automations.get(item_to_delete)

        if not existing_automation:
            continue

        report.deleted.append(existing_automation)
        try:
            await client.delete_automation(automation_id=existing_automation.id)
            log.info(f"{item_to_delete} Deleted")
        except PrefectHTTPStatusError as exc:
            if exc.response.status_code == 404:
                log.info(f"{item_to_delete} was already deleted")
            else:
                raise

    log.info(
        f"Processed {trigger_log_message}: {len(report.created)} created, {len(report.updated)} updated, "
        f"{len(report.refreshed)} refreshed, {len(report.unchanged)} unchanged, {len(report.deleted)} deleted"
    )

    return report


async def gather_all_automations(client: PrefectClient) -> list[Automation]:
    """Gather all automations from the Prefect server

    By default the Prefect client only retrieves a limited number of automations, this function
    retrieves them all by paginating through the results. The default within Prefect is 200 items,
    and client.read_automations() doesn't support pagination parameters.
    """
    offset = 0
    limit = 200
    automations: list[Automation] = []
    while True:
        response = await client.request("POST", "/automations/filter", json={"limit": limit, "offset": offset})
        response.raise_for_status()
        batch = Automation.model_validate_list(response.json())
        automations.extend(batch)
        if len(batch) < limit:
            break
        offset += limit

    return automations



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub_sdk.exceptions import Error

if TYPE_CHECKING:
    from .models import UndefinedJinja2Error


class JinjaTemplateError(Error):
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return str(self.message)


class JinjaTemplateNotFoundError(JinjaTemplateError):
    def __init__(self, message: str | None, filename: str, base_template: str | None = None) -> None:
        self.message = message or "Template Not Found"
        self.filename = filename
        self.base_template = base_template


class JinjaTemplateSyntaxError(JinjaTemplateError):
    def __init__(self, message: str | None, lineno: int, filename: str | None = None) -> None:
        self.message = message or "Syntax Error"
        self.filename = filename
        self.lineno = lineno


class JinjaTemplateUndefinedError(JinjaTemplateError):
    def __init__(self, message: str | None, errors: list[UndefinedJinja2Error]) -> None:
        self.message = message or "Undefined Error"
        self.errors = errors


class JinjaTemplateOperationViolationError(JinjaTemplateError):
    def __init__(self, message: str | None = None) -> None:
        self.message = message or "Forbidden code found in the template"



import linecache

from rich.syntax import Syntax
from rich.traceback import Frame, Traceback


def identify_faulty_jinja_code(traceback: Traceback, nbr_context_lines: int = 3) -> list[tuple[Frame, Syntax]]:
    """This function identifies the faulty Jinja2 code and beautify it to provide meaningful information to the user.

    We use the rich's Traceback to parse the complete stack trace and extract Frames for each exception found in the trace.
    """
    response = []

    # Extract only the Jinja related exception
    for frame in [frame for frame in traceback.trace.stacks[0].frames if frame.filename.endswith(".j2")]:
        code = "".join(linecache.getlines(frame.filename))
        lexer_name = Traceback._guess_lexer(frame.filename, code)
        syntax = Syntax(
            code,
            lexer_name,
            line_numbers=True,
            line_range=(frame.lineno - nbr_context_lines, frame.lineno + nbr_context_lines),
            highlight_lines={frame.lineno},
            code_width=88,
            theme=traceback.theme,
            dedent=False,
        )
        response.append((frame, syntax))

    return response



from __future__ import annotations

from prefect import task
from prefect.cache_policies import NONE

from infrahub.core.constants import RelationshipKind
from infrahub.core.registry import registry
from infrahub.database import InfrahubDatabase  # noqa: TC001  needed for prefect flow
from infrahub.workflows.catalogue import PROFILE_REFRESH_PROCESS

from .models import ProfileRefreshTriggerDefinition


@task(name="gather-trigger-profile-refresh", cache_policy=NONE)
async def gather_trigger_profile_refresh(
    db: InfrahubDatabase | None = None,  # noqa: ARG001 Needed to have a common function signature for gathering functions
) -> list[ProfileRefreshTriggerDefinition]:
    """Gather profile refresh triggers for all profile schemas.

    This function creates trigger definitions for each profile schema that will
    listen for `NodeUpdatedEvent` on profiles. When a profile's attributes or
    relationships change, the trigger will fire and execute the profile refresh
    workflow to re-apply profiles to all related nodes.
    """
    branches_with_diff_from_main = registry.get_altered_schema_branches()
    branches_to_process: list[tuple[str, list[str]]] = [(branch, []) for branch in branches_with_diff_from_main]
    branches_to_process.append((registry.default_branch, branches_with_diff_from_main))

    triggers: list[ProfileRefreshTriggerDefinition] = []

    for branch_scope, branches_out_of_scope in branches_to_process:
        schema_branch = registry.schema.get_schema_branch(name=branch_scope)

        for profile_name in schema_branch.profile_names:
            profile_schema = schema_branch.get_profile(name=profile_name, duplicate=False)

            trigger_attr = [attr.name for attr in profile_schema.attributes if attr.name != "profile_name"]
            trigger_rels = [
                rel.name
                for rel in profile_schema.relationships
                if rel.kind in {RelationshipKind.GENERIC, RelationshipKind.ATTRIBUTE} and rel.name != "related_nodes"
            ]
            trigger_fields = sorted(trigger_attr + trigger_rels)

            if trigger_fields:
                triggers.append(
                    ProfileRefreshTriggerDefinition.from_profile_schema(
                        branch=branch_scope,
                        profile_kind=profile_schema.kind,
                        trigger_fields=trigger_fields,
                        workflow=PROFILE_REFRESH_PROCESS,
                        branches_out_of_scope=branches_out_of_scope,
                    )
                )

    return triggers



from __future__ import annotations

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING

from .exceptions import UninitializedError
from .operation import InfrahubOperation

if TYPE_CHECKING:
    from .client import InfrahubClient
    from .context import RequestContext
    from .node import InfrahubNode


class InfrahubGenerator(InfrahubOperation):
    """Infrahub Generator class"""

    def __init__(
        self,
        query: str,
        client: InfrahubClient,
        infrahub_node: type[InfrahubNode],
        branch: str = "",
        root_directory: str = "",
        generator_instance: str = "",
        params: dict | None = None,
        convert_query_response: bool = False,
        execute_in_proposed_change: bool = True,
        execute_after_merge: bool = True,
        logger: logging.Logger | None = None,
        request_context: RequestContext | None = None,
    ) -> None:
        self.query = query

        super().__init__(
            client=client,
            infrahub_node=infrahub_node,
            convert_query_response=convert_query_response,
            branch=branch,
            root_directory=root_directory,
        )

        self.params = params or {}
        self.generator_instance = generator_instance
        self._client: InfrahubClient | None = None
        self.logger = logger or logging.getLogger("infrahub.tasks")
        self.request_context = request_context
        self.execute_in_proposed_change = execute_in_proposed_change
        self.execute_after_merge = execute_after_merge

    @property
    def subscribers(self) -> list[str] | None:
        if self.generator_instance:
            return [self.generator_instance]
        return None

    @property
    def client(self) -> InfrahubClient:
        if self._client:
            return self._client
        raise UninitializedError("The client has not been initialized")

    @client.setter
    def client(self, value: InfrahubClient) -> None:
        self._client = value

    async def collect_data(self) -> dict:
        """Query the result of the GraphQL Query defined in self.query and return the result"""

        data = await self._init_client.query_gql_query(
            name=self.query,
            branch_name=self.branch_name,
            variables=self.params,
            update_group=True,
            subscribers=self.subscribers,
        )
        return data.get("data") or data

    async def run(self, identifier: str, data: dict | None = None) -> None:
        """Execute the generator after collecting the data from the GraphQL query."""

        if not data:
            data = await self.collect_data()
        unpacked = data.get("data") or data
        await self.process_nodes(data=unpacked)

        group_type = "CoreGeneratorGroup" if self.execute_after_merge else "CoreGeneratorAwareGroup"

        async with self._init_client.start_tracking(
            identifier=identifier, params=self.params, delete_unused_nodes=True, group_type=group_type
        ) as self.client:
            await self.generate(data=unpacked)

    @abstractmethod
    async def generate(self, data: dict) -> None:
        """Code to run the generator

        Any child class of the InfrahubGenerator us expected to provide this method. The method is expected
        to use the provided InfrahubClient contained in self.client to create or update any nodes in an idempotent
        way as the method could be executed multiple times. Typically this would be done by using:

        await new_or_updated_object.save(allow_upsert=True)

        The tracking system will be responsible for deleting nodes that are no longer required.
        """



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from infrahub.core.query.node import NodeGetByHFIDQuery

from .queries.get_profile_data import GetProfileDataQuery, RelationshipFilter

if TYPE_CHECKING:
    from collections.abc import Sequence

    from infrahub.core.branch import Branch
    from infrahub.core.schema import NodeSchema
    from infrahub.database import InfrahubDatabase


@dataclass
class ProfileIdentifiers:
    ids: list[str]
    hfids: list[list[str]]


def _extract_profile_identifiers_from_input(profiles_data: Sequence[Any] | None) -> ProfileIdentifiers:
    """Extract profile IDs and HFIDs from input data."""
    if not profiles_data:
        return ProfileIdentifiers(ids=[], hfids=[])

    ids: list[str] = []
    hfids: list[list[str]] = []
    for item in profiles_data:
        if isinstance(item, str):
            # IDs
            ids.append(item)
        elif isinstance(item, dict):
            # Dicts with `id` or `hfid` keys
            if profile_id := item.get("id"):
                ids.append(profile_id)
            elif profile_hfid := item.get("hfid"):
                hfids.append(profile_hfid)
        elif hasattr(item, "id"):
            # Avoid circular import by not importing Node directly
            ids.append(item.id)

    return ProfileIdentifiers(ids=ids, hfids=hfids)


async def _resolve_hfids_to_ids(
    db: InfrahubDatabase, branch: Branch, profile_kind: str, hfids: list[list[str]]
) -> list[str]:
    query = await NodeGetByHFIDQuery.init(db=db, branch=branch, node_kind=profile_kind, hfids=hfids)
    await query.execute(db=db)
    return query.get_node_uuids()


class ProfilesMandatoryFieldGetter:
    def __init__(self, db: InfrahubDatabase, branch: Branch) -> None:
        self.db = db
        self.branch = branch

    async def get_mandatory_fields_from_profiles(
        self,
        schema: NodeSchema,
        profiles_data: Sequence[Any] | None,
        mandatory_attr_names: list[str],
        mandatory_rel_names: list[str],
    ) -> tuple[set[str], set[str]]:
        """Get mandatory attributes and relationships that are provided by profiles."""
        identifiers = _extract_profile_identifiers_from_input(profiles_data=profiles_data)

        profile_ids = list(identifiers.ids)
        if identifiers.hfids:
            resolved_ids = await _resolve_hfids_to_ids(
                db=self.db, branch=self.branch, profile_kind=f"Profile{schema.kind}", hfids=identifiers.hfids
            )
            profile_ids.extend(resolved_ids)

        if not profile_ids:
            return set(), set()

        rel_filters: list[RelationshipFilter] = []
        rel_name_to_filter: dict[str, RelationshipFilter] = {}
        for rel_name in mandatory_rel_names:
            rel_schema = schema.get_relationship(name=rel_name)
            if not rel_schema.support_profiles:
                continue

            rel_filter = RelationshipFilter(
                relationship_identifier=f"profile_{rel_schema.get_identifier()}", direction=rel_schema.direction
            )
            rel_filters.append(rel_filter)
            rel_name_to_filter[rel_name] = rel_filter

        query = await GetProfileDataQuery.init(
            db=self.db,
            branch=self.branch,
            profile_ids=profile_ids,
            attr_names=mandatory_attr_names,
            relationship_filters=rel_filters,
        )
        await query.execute(db=self.db)
        profile_data_list = query.get_profile_data()

        provided_attrs: set[str] = set()
        provided_rels: set[str] = set()

        for profile_data in profile_data_list:
            for attr_name in mandatory_attr_names:
                if profile_data.attribute_values.get(attr_name) is not None:
                    provided_attrs.add(attr_name)

            for rel_name, rel_filter in rel_name_to_filter.items():
                if profile_data.relationship_peers.get(rel_filter):
                    provided_rels.add(rel_name)

        return provided_attrs, provided_rels



import ast

from graphql import (
    FieldNode,
    FragmentDefinitionNode,
    FragmentSpreadNode,
    InlineFragmentNode,
    OperationDefinitionNode,
    SelectionNode,
    SelectionSetNode,
)


def strip_typename_from_selection_set(selection_set: SelectionSetNode | None) -> SelectionSetNode | None:
    """Recursively strip __typename fields from a SelectionSetNode.

    The __typename meta-field is an introspection field that is not part of the schema's
    type definitions. When code generation tools like ariadne-codegen try to look up
    __typename in the schema, they fail because it's a reserved introspection field.

    This function removes all __typename fields from the selection set, allowing
    code generation to proceed without errors.
    """
    if selection_set is None:
        return None

    new_selections: list[SelectionNode] = []
    for selection in selection_set.selections:
        if isinstance(selection, FieldNode):
            # Skip __typename fields
            if selection.name.value == "__typename":
                continue
            # Recursively process nested selection sets
            new_field = FieldNode(
                alias=selection.alias,
                name=selection.name,
                arguments=selection.arguments,
                directives=selection.directives,
                selection_set=strip_typename_from_selection_set(selection.selection_set),
            )
            new_selections.append(new_field)
        elif isinstance(selection, InlineFragmentNode):
            # Process inline fragments
            new_inline = InlineFragmentNode(
                type_condition=selection.type_condition,
                directives=selection.directives,
                selection_set=strip_typename_from_selection_set(selection.selection_set),
            )
            new_selections.append(new_inline)
        elif isinstance(selection, FragmentSpreadNode):
            # FragmentSpread references a named fragment - keep as-is
            new_selections.append(selection)
        else:
            raise TypeError(f"Unexpected GraphQL selection node type '{type(selection).__name__}'.")

    return SelectionSetNode(selections=tuple(new_selections))


def strip_typename_from_operation(operation: OperationDefinitionNode) -> OperationDefinitionNode:
    """Strip __typename fields from an operation definition.

    Returns a new OperationDefinitionNode with all __typename fields removed
    from its selection set and any nested selection sets.
    """
    return OperationDefinitionNode(
        operation=operation.operation,
        name=operation.name,
        variable_definitions=operation.variable_definitions,
        directives=operation.directives,
        selection_set=strip_typename_from_selection_set(operation.selection_set),
    )


def strip_typename_from_fragment(fragment: FragmentDefinitionNode) -> FragmentDefinitionNode:
    """Strip __typename fields from a fragment definition.

    Returns a new FragmentDefinitionNode with all __typename fields removed
    from its selection set and any nested selection sets.
    """
    return FragmentDefinitionNode(
        name=fragment.name,
        type_condition=fragment.type_condition,
        variable_definitions=fragment.variable_definitions,
        directives=fragment.directives,
        selection_set=strip_typename_from_selection_set(fragment.selection_set),
    )


def get_class_def_index(module: ast.Module) -> int:
    """Get the index of the first class definition in the module.
    It's useful to insert other classes before the first class definition."""
    for idx, item in enumerate(module.body):
        if isinstance(item, ast.ClassDef):
            return idx
    return -1


def insert_fragments_inline(module: ast.Module, fragment: ast.Module) -> ast.Module:
    """Insert the Pydantic classes for the fragments inline into the module.

    If no class definitions exist in module, fragments are appended to the end.
    """
    module_class_def_index = get_class_def_index(module)

    fragment_classes: list[ast.ClassDef] = [item for item in fragment.body if isinstance(item, ast.ClassDef)]

    # Handle edge case when no class definitions exist
    if module_class_def_index == -1:
        # Append fragments to the end of the module
        module.body.extend(fragment_classes)
    else:
        # Insert fragments before the first class definition
        for idx, item in enumerate(fragment_classes):
            module.body.insert(module_class_def_index + idx, item)

    return module


def remove_fragment_import(module: ast.Module) -> ast.Module:
    """Remove the fragment import from the module."""
    for item in module.body:
        if isinstance(item, ast.ImportFrom) and item.module == "fragments":
            module.body.remove(item)
            return module
    return module



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core.branch import Branch
from infrahub.core.constants import InfrahubKind, MetadataOptions
from infrahub.core.constants.schema import RESOURCE_POOL_REL_SUFFIX
from infrahub.core.initialization import get_root_node
from infrahub.core.manager import NodeManager
from infrahub.core.migrations.shared import (
    MigrationInput,
    MigrationRequiringRebase,
    MigrationResult,
    get_migration_console,
)

from .load_schema_branch import get_or_load_schema_branch

if TYPE_CHECKING:
    from infrahub.core.node import Node
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.database import InfrahubDatabase

console = get_migration_console()

IP_POOL_KINDS = {InfrahubKind.IPADDRESSPOOL, InfrahubKind.IPPREFIXPOOL}


class Migration064(MigrationRequiringRebase):
    """Migrate pool-sourced relationships and attributes on templates to _from_resource_pool relationships.

    Handles two cases:
    1. IP relationships sourced from IP pools → creates _from_resource_pool relationship, deletes original
    2. Number attributes sourced from Number pools → creates _from_resource_pool relationship, clears attribute source
    """

    name: str = "064_template_pool_relationship_cleanup"
    minimum_version: int = 63

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db
        root_node = await get_root_node(db=db, initialize=False)
        default_branch_name = root_node.default_branch
        default_branch = await Branch.get_by_name(db=db, name=default_branch_name)
        schema_branch = await get_or_load_schema_branch(db=db, branch=default_branch)
        return await self._process_templates(
            db=db, branch=default_branch, schema_branch=schema_branch, migration_input=migration_input
        )

    async def execute_against_branch(self, migration_input: MigrationInput, branch: Branch) -> MigrationResult:
        db = migration_input.db
        schema_branch = await get_or_load_schema_branch(db=db, branch=branch)
        return await self._process_templates(
            db=db, branch=branch, schema_branch=schema_branch, migration_input=migration_input
        )

    async def _process_templates(
        self, db: InfrahubDatabase, branch: Branch, schema_branch: SchemaBranch, migration_input: MigrationInput
    ) -> MigrationResult:
        try:
            for template_kind in schema_branch.template_names:
                template_schema = schema_branch.get(name=template_kind, duplicate=False)

                # Build map: original_rel_name -> pool_rel_name (for IP relationships)
                rel_pool_map: dict[str, str] = {}
                # Build map: attr_name -> pool_rel_name (for Number attributes)
                attr_pool_map: dict[str, str] = {}

                for rel in template_schema.relationships:
                    if not rel.name.endswith(RESOURCE_POOL_REL_SUFFIX):
                        continue
                    original_name = rel.name.removesuffix(RESOURCE_POOL_REL_SUFFIX)
                    if original_name in template_schema.relationship_names:
                        rel_pool_map[original_name] = rel.name
                    elif original_name in template_schema.attribute_names:
                        attr_pool_map[original_name] = rel.name

                if not rel_pool_map and not attr_pool_map:
                    continue

                templates = await NodeManager.query(
                    db=db,
                    schema=template_kind,
                    branch=branch,
                    include_metadata=MetadataOptions.LINKED_NODES,
                )

                for template in templates:
                    await self._process_one_template(
                        db=db,
                        template=template,
                        rel_pool_map=rel_pool_map,
                        attr_pool_map=attr_pool_map,
                        migration_input=migration_input,
                    )

        except Exception as exc:
            error_msg = str(exc) or f"{type(exc).__name__}: {repr(exc)}"
            return MigrationResult(errors=[error_msg])

        return MigrationResult()

    async def _process_one_template(
        self,
        db: InfrahubDatabase,
        template: Node,
        rel_pool_map: dict[str, str],
        attr_pool_map: dict[str, str],
        migration_input: MigrationInput,
    ) -> None:
        at = migration_input.at
        user_id = migration_input.user_id

        # Handle IP relationship pool sources
        for original_rel_name, pool_rel_name in rel_pool_map.items():
            rel_mgr = template.get_relationship(original_rel_name)

            # Find the first relationship sourced from an IP pool (at most one per manager)
            rels = await rel_mgr.get_relationships(db=db)
            for rel in rels:
                if not getattr(rel, "source_id", None):
                    continue
                source = await rel.get_source(db=db)
                if not source or source.get_kind() not in IP_POOL_KINDS:
                    continue

                # Create _from_resource_pool relationship pointing to the pool
                pool_rel_mgr = template.get_relationship(pool_rel_name)
                await pool_rel_mgr.update(db=db, data=source)
                await pool_rel_mgr.save(db=db, at=at, user_id=user_id)

                # Soft-delete the original pool-sourced relationship
                await rel.delete(db=db, at=at, user_id=user_id)
                break

        # Handle Number attribute pool sources
        for attr_name, pool_rel_name in attr_pool_map.items():
            attribute = template.get_attribute(attr_name)
            if not attribute or not getattr(attribute, "source_id", None):
                continue

            source = await attribute.get_source(db=db)
            if not source or source.get_kind() != InfrahubKind.NUMBERPOOL:
                continue

            # Check if _from_resource_pool relationship already exists (idempotency)
            pool_rel_mgr = template.get_relationship(pool_rel_name)
            existing_pool_rels = await pool_rel_mgr.get_relationships(db=db)

            if not existing_pool_rels or existing_pool_rels[0].peer_id != source.id:
                # Create or repair the _from_resource_pool relationship
                await pool_rel_mgr.update(db=db, data=source)
                await pool_rel_mgr.save(db=db, at=at, user_id=user_id)

            # Clear the attribute source
            attribute.clear_source()
            await attribute.save(db=db, at=at, user_id=user_id)



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from .m001_add_version_to_graph import Migration001
from .m002_attribute_is_default import Migration002
from .m003_relationship_parent_optional import Migration003
from .m004_add_attr_documentation import Migration004
from .m005_add_rel_read_only import Migration005
from .m006_add_rel_on_delete import Migration006
from .m007_add_rel_allow_override import Migration007
from .m008_add_human_friendly_id import Migration008
from .m009_add_generate_profile_attr import Migration009
from .m010_add_generate_profile_attr_generic import Migration010
from .m011_remove_profile_relationship_schema import Migration011
from .m012_convert_account_generic import Migration012
from .m013_convert_git_password_credential import Migration013
from .m014_remove_index_attr_value import Migration014
from .m015_diff_format_update import Migration015
from .m016_diff_delete_bug_fix import Migration016
from .m017_add_core_profile import Migration017
from .m018_uniqueness_nulls import Migration018
from .m019_restore_rels_to_time import Migration019
from .m020_duplicate_edges import Migration020
from .m021_missing_hierarchy_merge import Migration021
from .m022_add_generate_template_attr import Migration022
from .m023_deduplicate_cardinality_one_relationships import Migration023
from .m024_missing_hierarchy_backfill import Migration024
from .m025_uniqueness_nulls import Migration025
from .m026_0000_prefix_fix import Migration026
from .m027_delete_isolated_nodes import Migration027
from .m028_delete_diffs import Migration028
from .m029_duplicates_cleanup import Migration029
from .m030_illegal_edges import Migration030
from .m031_check_number_attributes import Migration031
from .m032_cleanup_orphaned_branch_relationships import Migration032
from .m033_deduplicate_relationship_vertices import Migration033
from .m034_find_orphaned_schema_fields import Migration034
from .m035_orphan_relationships import Migration035
from .m036_drop_attr_value_index import Migration036
from .m037_index_attr_vals import Migration037
from .m038_redo_0000_prefix_fix import Migration038
from .m039_ipam_reconcile import Migration039
from .m040_duplicated_attributes import Migration040
from .m041_deleted_dup_edges import Migration041
from .m042_profile_attrs_in_db import Migration042
from .m043_create_hfid_display_label_in_db import Migration043
from .m044_backfill_hfid_display_label_in_db import Migration044
from .m045_backfill_hfid_display_label_in_db_profile_template import Migration045
from .m046_fill_agnostic_hfid_display_labels import Migration046
from .m047_backfill_or_null_display_label import Migration047
from .m048_undelete_rel_props import Migration048
from .m049_remove_is_visible_relationship import Migration049
from .m050_backfill_vertex_metadata import Migration050
from .m051_subtract_branched_from_microsecond import Migration051
from .m052_fix_global_branch_level import Migration052
from .m053_fix_branch_level_zero import Migration053
from .m054_cleanup_orphaned_nodes import Migration054
from .m055_remove_webhook_validate_certificates_default import Migration055
from .m056_update_schema_node_generic_constraints import Migration056
from .m057_deduplicate_schema_nodes import Migration057
from .m058_remove_profiles_schema_relationships import Migration058
from .m059_fix_hfid_display_label_nulls import Migration059
from .m060_set_coreproposedchange_created_by_on_node import Migration060
from .m061_link_proposed_changes_to_diff_roots import Migration061
from .m062_recompute_permission_display_labels import Migration062
from .m063_template_number_pool_cleanup import Migration063
from .m064_template_ip_pool_relationship_cleanup import Migration064
from .m065_remove_generic_generate_template import Migration065
from .m066_consolidate_duplicate_number_pools import Migration066
from .m067_freeze_orphaned_branch_tracking_diffs import Migration067

if TYPE_CHECKING:
    from ..shared import MigrationTypes


MIGRATIONS: list[type[MigrationTypes]] = [
    Migration001,
    Migration002,
    Migration003,
    Migration004,
    Migration005,
    Migration006,
    Migration007,
    Migration008,
    Migration009,
    Migration010,
    Migration011,
    Migration012,
    Migration013,
    Migration014,
    Migration015,
    Migration016,
    Migration017,
    Migration018,
    Migration019,
    Migration020,
    Migration021,
    Migration022,
    Migration023,
    Migration024,
    Migration025,
    Migration026,
    Migration027,
    Migration028,
    Migration029,
    Migration030,
    Migration031,
    Migration032,
    Migration033,
    Migration034,
    Migration035,
    Migration036,
    Migration037,
    Migration038,
    Migration039,
    Migration040,
    Migration041,
    Migration042,
    Migration043,
    Migration044,
    Migration045,
    Migration046,
    Migration047,
    Migration048,
    Migration049,
    Migration050,
    Migration051,
    Migration052,
    Migration053,
    Migration054,
    Migration055,
    Migration056,
    Migration057,
    Migration058,
    Migration059,
    Migration060,
    Migration061,
    Migration062,
    Migration063,
    Migration064,
    Migration065,
    Migration066,
    Migration067,
]


async def get_graph_migrations(current_graph_version: int) -> Sequence[MigrationTypes]:
    applicable_migrations = []
    for migration_class in MIGRATIONS:
        migration = migration_class.init()
        if current_graph_version > migration.minimum_version:
            continue
        applicable_migrations.append(migration)

    return applicable_migrations


def get_migration_by_number(migration_number: int | str) -> MigrationTypes:
    # Convert to string and pad with zeros if needed
    try:
        num = int(migration_number)
        migration_str = f"{num:03d}"
    except (ValueError, TypeError) as exc:
        raise ValueError(f"Invalid migration number: {migration_number}") from exc

    migration_name = f"Migration{migration_str}"

    # Find the migration in the MIGRATIONS list
    for migration_class in MIGRATIONS:
        if migration_class.__name__ == migration_name:
            return migration_class.init()

    raise ValueError(f"Migration {migration_number} not found")



from infrahub.core.constants import (
    AllowOverrideType,
    BranchSupportType,
    GlobalPermissions,
    InfrahubKind,
    PermissionAction,
    PermissionDecision,
)
from infrahub.core.constants import RelationshipCardinality as Cardinality
from infrahub.core.constants import RelationshipKind as RelKind

from ...attribute_schema import AttributeSchema as Attr
from ...dropdown import DropdownChoice
from ...generic_schema import GenericSchema
from ...node_schema import NodeSchema
from ...relationship_schema import (
    RelationshipSchema as Rel,
)

core_base_permission = GenericSchema(
    name="BasePermission",
    namespace="Core",
    description="A permission grants right to an account",
    label="Base permission",
    icon="mdi:user-key",
    include_in_menu=False,
    generate_profile=False,
    attributes=[
        Attr(name="description", kind="Text", optional=True),
        Attr(
            name="identifier",
            kind="Text",
            description="Identifier for the permission",
            read_only=True,
            optional=True,
            allow_override=AllowOverrideType.NONE,
        ),
    ],
    relationships=[
        Rel(
            name="roles",
            peer=InfrahubKind.ACCOUNTROLE,
            optional=True,
            identifier="role__permissions",
            cardinality=Cardinality.MANY,
            kind=RelKind.ATTRIBUTE,
        )
    ],
)

core_object_permission = NodeSchema(
    name="ObjectPermission",
    namespace="Core",
    description="A permission that grants rights to perform actions on objects",
    label="Object permission",
    include_in_menu=False,
    order_by=["namespace__value", "name__value", "action__value", "decision__value"],
    display_label=(
        "object:{{ namespace__value }}:{{ name__value }}:{{ action__value | value_to_permission_action_name | lower }}:"
        "{{ decision__value | value_to_permission_decision_name | lower }}"
    ),
    human_friendly_id=["namespace__value", "name__value", "action__value", "decision__value"],
    uniqueness_constraints=[["namespace__value", "name__value", "action__value", "decision__value"]],
    generate_profile=False,
    inherit_from=[InfrahubKind.BASEPERMISSION],
    attributes=[
        Attr(
            name="namespace",
            kind="Text",
            description="Namespace of the object type this permission applies to",
            order_weight=2000,
        ),
        Attr(name="name", kind="Text", order_weight=3000),
        Attr(
            name="action",
            kind="Text",
            description="The action this permission grants or denies",
            enum=PermissionAction.available_types(),
            default_value=PermissionAction.ANY.value,
            order_weight=4000,
        ),
        Attr(
            name="decision",
            kind="Number",
            enum=PermissionDecision.available_types(),
            default_value=PermissionDecision.ALLOW_ALL.value,
            order_weight=5000,
            description="Decide to deny or allow the action."
            "If allowed, it can be configured for the default branch, any other branches or all branches",
        ),
    ],
)

core_global_permission = NodeSchema(
    name="GlobalPermission",
    namespace="Core",
    description="A permission that grants global rights to perform actions in Infrahub",
    label="Global permission",
    include_in_menu=False,
    order_by=["action__value", "decision__value"],
    display_label=(
        "global:{{ action__value | lower }}:{{ decision__value | value_to_permission_decision_name | lower }}"
    ),
    human_friendly_id=["action__value", "decision__value"],
    generate_profile=False,
    inherit_from=[InfrahubKind.BASEPERMISSION],
    branch=BranchSupportType.AGNOSTIC,
    attributes=[
        Attr(
            name="action",
            kind="Dropdown",
            description="The global action this permission grants or denies",
            choices=[DropdownChoice(name=permission.value) for permission in GlobalPermissions],
            order_weight=2000,
        ),
        Attr(
            name="decision",
            kind="Number",
            enum=PermissionDecision.available_types(),
            default_value=PermissionDecision.ALLOW_ALL.value,
            order_weight=3000,
            description="Decide to deny or allow the action at a global level",
        ),
    ],
)

core_account_role = NodeSchema(
    name="AccountRole",
    namespace="Core",
    description="A role defines a set of permissions to grant to a group of accounts",
    label="Account role",
    icon="mdi:user-badge",
    include_in_menu=False,
    order_by=["name__value"],
    display_labels=["name__value"],
    human_friendly_id=["name__value"],
    generate_profile=False,
    attributes=[Attr(name="name", kind="Text", unique=True)],
    relationships=[
        Rel(
            name="groups",
            peer=InfrahubKind.ACCOUNTGROUP,
            optional=True,
            identifier="role__accountgroups",
            cardinality=Cardinality.MANY,
            kind=RelKind.ATTRIBUTE,
        ),
        Rel(
            name="permissions",
            peer=InfrahubKind.BASEPERMISSION,
            optional=True,
            identifier="role__permissions",
            cardinality=Cardinality.MANY,
            kind=RelKind.ATTRIBUTE,
        ),
    ],
)

core_account_group = NodeSchema(
    name="AccountGroup",
    namespace="Core",
    description="A group of users to manage common permissions",
    label="Account group",
    icon="mdi:account-group",
    include_in_menu=False,
    order_by=["name__value"],
    display_labels=["name__value"],
    human_friendly_id=["name__value"],
    generate_profile=False,
    inherit_from=[InfrahubKind.LINEAGEOWNER, InfrahubKind.LINEAGESOURCE, InfrahubKind.GENERICGROUP],
    branch=BranchSupportType.AGNOSTIC,
    relationships=[
        Rel(
            name="roles",
            peer=InfrahubKind.ACCOUNTROLE,
            optional=True,
            identifier="role__accountgroups",
            cardinality=Cardinality.MANY,
            kind=RelKind.ATTRIBUTE,
        ),
    ],
)



from enum import StrEnum


class DatabaseType(StrEnum):
    NEO4J = "neo4j"
    MEMGRAPH = "memgraph"


class Neo4jRuntime(StrEnum):
    DEFAULT = "default"
    INTERPRETED = "interpreted"
    SLOTTED = "slotted"
    PIPELINED = "pipelined"
    PARALLEL = "parallel"
    UNDEFINED = "undefined"


class IndexType(StrEnum):
    TEXT = "text"
    RANGE = "range"
    LOOKUP = "lookup"
    NOT_APPLICABLE = "not_applicable"


class EntityType(StrEnum):
    NODE = "node"
    RELATIONSHIP = "relationship"



import re
from typing import Any

import graphene

from infrahub.core.enums import generate_python_enum
from infrahub.core.schema import AttributeSchema, MainSchemaTypes

ENUM_NAME_REGEX = re.compile(r"[_a-zA-Z0-9]+")


def get_enum_attribute_type_name(node_schema: MainSchemaTypes, attr_schema: AttributeSchema) -> str:
    return f"{node_schema.kind}{attr_schema.name.title()}"


def generate_graphql_enum(name: str, options: list[Any]) -> graphene.Enum:
    def description_func(value: Any) -> str:
        if value:
            return value.value
        return f"Enum for {name}"

    py_enum = generate_python_enum(name=name, options=options)
    return graphene.Enum.from_enum(enum=py_enum, description=description_func)


class ConflictSelection(graphene.Enum):
    BASE_BRANCH = "base"
    DIFF_BRANCH = "diff"



from abc import ABC, abstractmethod

from infrahub.database import InfrahubDatabase

from ..models import PatchPlan


class PatchQuery(ABC):
    def __init__(self, db: InfrahubDatabase) -> None:
        self.db = db

    @abstractmethod
    async def plan(self) -> PatchPlan: ...

    @property
    @abstractmethod
    def name(self) -> str: ...



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from infrahub_sdk.template import Jinja2Template

from infrahub.core.constants import PermissionAction, PermissionDecision

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

__all__ = ["InfrahubJinja2Template"]


def _value_to_permission_decision_name(value: int | str | Enum) -> str:
    """Convert a permission decision value to its enum member name.

    Usage example: `{{ decision__value | value_to_permission_decision_name }}` will return `"ALLOW_ALL"` for value `6`.
    """
    raw_value = value.value if isinstance(value, Enum) else value

    try:
        int_value = int(raw_value)
    except ValueError as exc:
        msg = f"Value '{value}' is not a valid int for permission decision value: {exc}"
        raise ValueError(msg) from exc

    try:
        return PermissionDecision(int_value).name
    except ValueError as exc:
        msg = f"Value '{int_value}' not found in enum 'PermissionDecision': {exc}"
        raise ValueError(msg) from exc


def _value_to_permission_action_name(value: str | Enum) -> str:
    """Convert a permission action value to its enum member name.

    Usage examples:
    - `{{ action__value | value_to_permission_action_name }}` will return `"CREATE"` for value `"create"` for object permissions.
    - `{{ action__value | value_to_permission_action_name }}` will return `"SUPER_ADMIN"` for value `"super_admin"` for global permissions.
    """
    if isinstance(value, Enum):
        value = value.value

    try:
        return PermissionAction(str(value)).name
    except ValueError as exc:
        msg = f"Value '{value}' not found in enum 'PermissionAction': {exc}"
        raise ValueError(msg) from exc


FILTERS: dict[str, Callable[..., str]] = {
    "value_to_permission_decision_name": _value_to_permission_decision_name,
    "value_to_permission_action_name": _value_to_permission_action_name,
}


class InfrahubJinja2Template(Jinja2Template):
    """Extend SDK's `Jinja2Template` with Infrahub server-specific filters pre-configured."""

    def __init__(
        self,
        template: str | Path,
        template_directory: Path | None = None,
        filters: dict[str, Callable[..., str]] | None = None,
    ) -> None:
        super().__init__(
            template=template, template_directory=template_directory, filters={**FILTERS, **(filters or {})}
        )



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Literal, overload

from infrahub.core.constants import SYSTEM_USER_ID
from infrahub.core.migrations.schema.models import SchemaApplyMigrationData
from infrahub.core.migrations.schema.tasks import schema_apply_migrations
from infrahub.core.query.rollback import RollbackQuery
from infrahub.exceptions import MigrationError
from infrahub.log import get_logger
from infrahub.workflows.catalogue import SCHEMA_APPLY_MIGRATION

if TYPE_CHECKING:
    import logging

    from infrahub.context import InfrahubContext
    from infrahub.core.branch import Branch
    from infrahub.core.models import SchemaDiff, SchemaUpdateMigrationInfo
    from infrahub.core.schema.manager import SchemaManager
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase
    from infrahub.services.adapters.workflow import InfrahubWorkflow


_default_log = get_logger()


class MigrationExecutor(Enum):
    """Determines how migrations are executed."""

    DIRECT = "direct"  # Call schema_apply_migrations directly
    WORKFLOW = "workflow"  # Execute via workflow service


class SchemaUpdateCoordinator:
    """Coordinates schema updates and migrations with unified timestamps and rollback on failure.

    This class encapsulates the logic for:
    1. Updating schema in the database with a unified timestamp
    2. Running migrations (either directly or via workflow service)
    3. Rolling back all changes on failure
    4. Restoring registry state and branch hash
    """

    def __init__(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        schema_manager: SchemaManager,
        origin_schema: SchemaBranch,
        workflow: InfrahubWorkflow | None = None,
        context: InfrahubContext | None = None,
        migration_executor: MigrationExecutor = MigrationExecutor.WORKFLOW,
        logger: logging.Logger | logging.LoggerAdapter[logging.Logger] | None = None,
    ) -> None:
        """Initialize the coordinator.

        Args:
            db: Database connection
            branch: Branch being updated
            schema_manager: Schema manager for updating schema in DB and registry
            origin_schema: Original schema before update (for rollback)
            workflow: Workflow service for executing migrations (required for WORKFLOW executor)
            context: Infrahub context (required for WORKFLOW executor)
            migration_executor: How to execute migrations (DIRECT or WORKFLOW)
            logger: Logger to use (defaults to module logger)
        """
        self.db = db
        self.branch = branch
        self.schema_manager = schema_manager
        self.origin_schema = origin_schema
        self.workflow = workflow
        self.context = context
        self.migration_executor = migration_executor
        self.log = logger or _default_log

        if self.migration_executor is MigrationExecutor.WORKFLOW and (self.workflow is None or self.context is None):
            raise RuntimeError("Workflow and context are required for WORKFLOW executor")

    def _get_workflow(self) -> InfrahubWorkflow:
        """Get workflow service, raising if not available."""
        if self.workflow is None:
            raise RuntimeError("Workflow service is required but not provided")
        return self.workflow

    def _get_context(self) -> InfrahubContext:
        """Get context, raising if not available."""
        if self.context is None:
            raise RuntimeError("Context is required but not provided")
        return self.context

    @overload
    async def execute(
        self,
        candidate_schema: SchemaBranch,
        at: Timestamp,
        diff: SchemaDiff | None = None,
        migrations: list[SchemaUpdateMigrationInfo] | None = None,
        limit: list[str] | None = None,
        update_db: Literal[True] = True,
        update_registry: bool = True,
        user_id: str = SYSTEM_USER_ID,
    ) -> str: ...

    @overload
    async def execute(
        self,
        candidate_schema: SchemaBranch,
        at: Timestamp,
        diff: SchemaDiff | None = None,
        migrations: list[SchemaUpdateMigrationInfo] | None = None,
        limit: list[str] | None = None,
        update_db: Literal[False] = ...,
        update_registry: Literal[True] = ...,
        user_id: str = SYSTEM_USER_ID,
    ) -> str: ...

    @overload
    async def execute(
        self,
        candidate_schema: SchemaBranch,
        at: Timestamp,
        diff: SchemaDiff | None = None,
        migrations: list[SchemaUpdateMigrationInfo] | None = None,
        limit: list[str] | None = None,
        update_db: Literal[False] = ...,
        update_registry: Literal[False] = ...,
        user_id: str = SYSTEM_USER_ID,
    ) -> str | None: ...

    async def execute(
        self,
        candidate_schema: SchemaBranch,
        at: Timestamp,
        diff: SchemaDiff | None = None,
        migrations: list[SchemaUpdateMigrationInfo] | None = None,
        limit: list[str] | None = None,
        update_db: bool = True,
        update_registry: bool = True,
        user_id: str = SYSTEM_USER_ID,
    ) -> str | None:
        """Execute the schema update with migrations and rollback on failure.

        Args:
            candidate_schema: New schema to apply
            at: Timestamp for all operations (enables atomic rollback)
            diff: Schema diff for update_schema_branch (optional, for incremental updates)
            migrations: List of migrations to run
            limit: Limit schema update to specific items
            update_db: Whether to update the database
            update_registry: Whether to update the registry (used when DB already updated externally)
            user_id: User ID for all operations

        Returns:
            The updated schema hash, or None if the schema was not updated

        Raises:
            MigrationError: If migrations fail and rollback completes
            Exception: Original exception if migrations fail via exception
        """

        # Step 1: Update schema in DB and/or registry
        updated_hash: str | None = None
        if update_db or update_registry:
            updated_hash = await self._update_schema(
                candidate_schema=candidate_schema,
                at=at,
                diff=diff,
                limit=limit,
                update_db=update_db,
                update_registry=update_registry,
                user_id=user_id,
            )

        if not migrations:
            return updated_hash

        # Step 2: Run migrations
        error_msgs, exception = await self._run_migrations(
            candidate_schema=candidate_schema,
            at=at,
            migrations=migrations,
            user_id=user_id,
        )

        # Step 3: Check for failures
        if not error_msgs and exception is None:
            return updated_hash

        if exception:
            self.log.error(
                "Schema migration failed, beginning rollback",
                extra={"branch": self.branch.name, "error": str(exception)},
            )
        else:
            self.log.error(
                "Schema migration returned errors, beginning rollback",
                extra={"branch": self.branch.name, "errors": error_msgs},
            )

        # Step 4: Rollback on failure
        await self._rollback(at=at)
        await self._restore_registry_state()

        self.log.info("Schema rollback completed", extra={"branch": self.branch.name})

        if exception:
            raise exception
        raise MigrationError(message=",\n".join(error_msgs))

    async def _update_schema(
        self,
        candidate_schema: SchemaBranch,
        at: Timestamp,
        diff: SchemaDiff | None,
        limit: list[str] | None,
        update_db: bool,
        update_registry: bool,
        user_id: str,
    ) -> str:
        """Update schema in database and/or registry.

        Args:
            candidate_schema: The new schema to apply
            at: Timestamp for DB operations
            diff: Schema diff for incremental updates
            limit: Limit to specific schema items
            update_db: If True, update DB via update_schema_branch.
            update_registry: If True, update registry via set_schema_branch.
        Returns:
            The updated schema hash
        """
        if update_db:
            await self.schema_manager.update_schema_branch(
                schema=candidate_schema,
                db=self.db,
                branch=self.branch.name,
                diff=diff,
                limit=limit,
                update_db=True,
                at=at,
                user_id=user_id,
            )
        if update_registry:
            self.schema_manager.set_schema_branch(name=self.branch.name, schema=candidate_schema)

        self.branch.update_schema_hash()
        self.log.info(
            "Schema has been updated", extra={"branch": self.branch.name, "hash": self.branch.active_schema_hash.main}
        )
        await self.branch.save(db=self.db, user_id=user_id)

        return candidate_schema.get_hash()

    async def _run_migrations(
        self,
        candidate_schema: SchemaBranch,
        at: Timestamp,
        migrations: list[SchemaUpdateMigrationInfo],
        user_id: str,
    ) -> tuple[list[str], Exception | None]:
        """Run migrations using the configured executor."""
        apply_migration_data = SchemaApplyMigrationData(
            branch=self.branch,
            new_schema=candidate_schema,
            previous_schema=self.origin_schema,
            migrations=migrations,
            at=at,
            user_id=user_id,
        )

        if self.migration_executor == MigrationExecutor.WORKFLOW:
            return await self._run_migrations_via_workflow(apply_migration_data=apply_migration_data)
        return await self._run_migrations_directly(apply_migration_data=apply_migration_data)

    async def _run_migrations_via_workflow(
        self,
        apply_migration_data: SchemaApplyMigrationData,
    ) -> tuple[list[str], Exception | None]:
        """Execute migrations via workflow service."""
        workflow = self._get_workflow()
        context = self._get_context()

        error_msgs: list[str] = []
        exception: Exception | None = None

        try:
            error_msgs = await workflow.execute_workflow(
                workflow=SCHEMA_APPLY_MIGRATION,
                context=context,
                expected_return=list[str],
                parameters={"message": apply_migration_data},
            )
        except Exception as exc:
            exception = exc

        return error_msgs, exception

    async def _run_migrations_directly(
        self,
        apply_migration_data: SchemaApplyMigrationData,
    ) -> tuple[list[str], Exception | None]:
        """Execute migrations directly (for branch tasks)."""

        error_msgs: list[str] = []
        exception: Exception | None = None

        try:
            error_msgs = await schema_apply_migrations(message=apply_migration_data)
        except Exception as exc:
            exception = exc

        return error_msgs, exception

    async def _rollback(self, at: Timestamp) -> None:
        """Rollback all changes made at the unified timestamp."""
        rollback_query = await RollbackQuery.init(db=self.db, target_branch=self.branch, at=at)
        await rollback_query.execute(db=self.db)

    async def _restore_registry_state(self) -> None:
        """Restore original schema in registry and reset branch hash."""
        self.schema_manager.set_schema_branch(name=self.branch.name, schema=self.origin_schema)
        self.branch.update_schema_hash()
        await self.branch.save(db=self.db)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Scalar
from graphql import GraphQLError

if TYPE_CHECKING:
    from graphql.language import ast
    from starlette.datastructures import UploadFile


class Upload(Scalar):
    """GraphQL scalar for file uploads.

    This scalar type handles file uploads in `multipart/form-data requests` following the GraphQL Multipart Request specification.

    The `Upload` scalar is input-only and cannot be used in query responses.
    """

    @staticmethod
    def serialize(_value: object) -> None:
        """Serialize is not supported for `Upload` scalar.

        Upload is an input-only type and cannot be returned in query responses.
        """
        raise GraphQLError("Upload scalar cannot be serialized. It is input-only.")

    @staticmethod
    def parse_value(value: UploadFile | None) -> UploadFile | None:
        """Parse the value from multipart form data.

        Args:
            value: A `UploadFile` object from Starlette's multipart parsing, or `None` if no file was provided.

        Returns:
            A Starlette's `UploadFile` object, or `None` if no file was provided.
        """
        return value

    @staticmethod
    def parse_literal(_node: ast.Node, _variables: dict[str, Any] | None = None) -> None:
        """Parse literal values is not supported for `Upload` scalar.

        `Upload` values must be provided via multipart form data, not as literal values in the GraphQL query.
        """
        raise GraphQLError("Upload scalar cannot be used as a literal value. Use multipart form data.")



import typer
from infrahub_sdk.async_typer import AsyncTyper

from infrahub import config
from infrahub.core.initialization import initialization

from ..workers.dependencies import get_database
from .context import CliContext
from .db import app as db_app
from .dev import app as dev_app
from .events import app as events_app
from .git_agent import app as git_app
from .server import app as server_app
from .tasks import app as tasks_app
from .upgrade import upgrade_cmd

app = AsyncTyper(name="Infrahub CLI", pretty_exceptions_enable=False)


@app.callback()
def common(ctx: typer.Context) -> None:
    """Infrahub CLI"""
    ctx.obj = CliContext()


app.add_typer(server_app, name="server")
app.add_typer(git_app, name="git-agent", hidden=True)
app.add_typer(db_app, name="db")
app.add_typer(events_app, name="events", help="Interact with the events system.", hidden=True)
app.add_typer(tasks_app, name="tasks", hidden=True)
app.add_typer(dev_app, name="dev", help="Internal development commands.")
app.command(name="upgrade")(upgrade_cmd)


async def _init_shell(config_file: str) -> None:
    """Launch a Python Interactive shell."""
    config.load_and_exit(config_file_name=config_file)

    db = await get_database()

    async with db.start_session() as db:
        await initialization(db=db)


@app.command()
def shell() -> None:
    """Start a python shell within Infrahub context (requires IPython)."""
    from infrahub_sdk import InfrahubClient
    from IPython import start_ipython
    from traitlets.config import Config

    from infrahub import config
    from infrahub.components import ComponentType
    from infrahub.core.branch import Branch
    from infrahub.core.initialization import initialization
    from infrahub.core.manager import NodeManager
    from infrahub.core.registry import registry
    from infrahub.dependencies.registry import build_component_registry
    from infrahub.lock import initialize_lock
    from infrahub.services import InfrahubServices
    from infrahub.workers.dependencies import (
        get_cache,
        get_component,
        get_database,
        get_workflow,
        set_component_type,
    )

    async def initialize_service() -> InfrahubServices:
        config.load_and_exit()
        client = InfrahubClient()

        component_type = ComponentType.GIT_AGENT
        set_component_type(component_type=component_type)

        database = await get_database()

        build_component_registry()

        workflow = get_workflow()
        cache = await get_cache()
        component = await get_component()
        service = await InfrahubServices.new(
            cache=cache,
            client=client,
            database=database,
            workflow=workflow,
            component=component,
            component_type=component_type,
        )
        initialize_lock(service=service)

        async with service.database as db:
            await initialization(db=db)
        await service.component.refresh_schema_hash()

        return service

    def welcome() -> None:
        print("--------------------------------------")
        print("infrahub interactive shell initialized")
        print("--------------------------------------")
        print("Available objects:")
        print("* db: InfrahubDatabase")
        print("* Branch: Branch")
        print("* NodeManager: NodeManager")
        print("* registry: Registry")
        print("* service: InfrahubServices")
        print()
        print("Example use:")
        print("In [1] tags = await NodeManager.query(schema='BuiltinTag', db=db)")

    c = Config()
    c.InteractiveShellApp.exec_lines = [
        "service = await initialize_service()",
        "db = service.database",
        "welcome()",
    ]
    c.TerminalInteractiveShell.colors = "Neutral"

    user_ns = {
        "initialize_service": initialize_service,
        "welcome": welcome,
        "NodeManager": NodeManager,
        "Branch": Branch,
        "registry": registry,
    }

    start_ipython(argv=[], config=c, user_ns=user_ns)



from __future__ import annotations

import ipaddress
from typing import TYPE_CHECKING, Iterable

from infrahub.core.query.ipam import (
    IPAddressData,
    IPPrefixData,
    IPPrefixIPAddressFetch,
    IPPrefixIPAddressFetchFree,
    IPPrefixSubnetFetch,
    IPPrefixSubnetFetchFree,
    IPv6PrefixIPAddressFetchFree,
    IPv6PrefixSubnetFetchFree,
)

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.ipam.constants import IPAddressType, IPNetworkType
    from infrahub.core.node import Node
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


class IPAMResourceAllocator:
    """Allocator for IPAM resources (prefixes and addresses) within pools.

    This class provides optimized methods for finding the next available
    IP prefix or address within a parent prefix. It uses database-side
    Cypher computation for performance, which is especially important
    for IPv6 allocations where the address space is very large.
    """

    def __init__(
        self,
        db: InfrahubDatabase,
        namespace: Node | str | None = None,
        branch: Branch | None = None,
        branch_agnostic: bool = False,
    ) -> None:
        """Initialize the IPAM resource allocator.

        Args:
            db: Database connection to use for queries.
            namespace: IP namespace node or its ID. If None, uses the default namespace.
            branch: Branch to query. If None, uses the default branch.
            branch_agnostic: If True, queries across all branches.
        """
        self.db = db
        self.namespace = namespace
        self.branch = branch
        self.branch_agnostic = branch_agnostic

    async def _get_next_ipv4_prefix(
        self, ip_prefix: IPNetworkType, target_prefix_length: int, at: Timestamp | str | None = None
    ) -> IPNetworkType | None:
        """Get the next available free IPv4 prefix.

        Uses integer arithmetic in Cypher for efficient gap detection in the
        IPv4 address space.

        Args:
            ip_prefix: Parent prefix to allocate from.
            target_prefix_length: Desired prefix length for the new prefix (0-32).
            at: Optional timestamp for point-in-time queries.

        Returns:
            The next available IPv4 prefix, or None if no space is available
            or if the target_prefix_length is invalid.
        """
        if target_prefix_length < 0 or target_prefix_length > 32:
            return None
        if target_prefix_length < ip_prefix.prefixlen:
            return None

        query = await IPPrefixSubnetFetchFree.init(
            db=self.db,
            branch=self.branch,
            obj=ip_prefix,
            target_prefixlen=target_prefix_length,
            namespace=self.namespace,
            at=at,
            branch_agnostic=self.branch_agnostic,
        )
        await query.execute(db=self.db)

        result_data = query.get_prefix_data()
        if not result_data:
            return None

        network_address = ipaddress.IPv4Address(result_data.free_start)
        return ipaddress.ip_network(f"{network_address}/{target_prefix_length}")

    async def _get_next_ipv6_prefix(
        self, ip_prefix: IPNetworkType, target_prefix_length: int, at: Timestamp | str | None = None
    ) -> IPNetworkType | None:
        """Get the next available free IPv6 prefix.

        Uses binary string operations in Cypher to handle IPv6's 128-bit address
        space, as integer values would overflow Neo4j's 64-bit integer type.

        Args:
            ip_prefix: Parent prefix to allocate from.
            target_prefix_length: Desired prefix length for the new prefix (0-128).
            at: Optional timestamp for point-in-time queries.

        Returns:
            The next available IPv6 prefix, or None if no space is available
            or if the target_prefix_length is invalid.
        """
        if target_prefix_length < 0 or target_prefix_length > 128:
            return None
        if target_prefix_length < ip_prefix.prefixlen:
            return None

        query = await IPv6PrefixSubnetFetchFree.init(
            db=self.db,
            branch=self.branch,
            obj=ip_prefix,
            target_prefixlen=target_prefix_length,
            namespace=self.namespace,
            at=at,
            branch_agnostic=self.branch_agnostic,
        )
        await query.execute(db=self.db)

        result_data = query.get_prefix_data()
        if not result_data:
            return None

        # Convert binary string to IPv6 address
        addr_int = int(result_data.free_start_bin, 2)
        network_address = ipaddress.IPv6Address(addr_int)
        return ipaddress.ip_network(f"{network_address}/{target_prefix_length}")

    async def get_next_prefix(
        self, ip_prefix: IPNetworkType, target_prefix_length: int, at: Timestamp | str | None = None
    ) -> IPNetworkType | None:
        """Get the next available free prefix of specified length within a parent prefix.

        Automatically selects the appropriate method based on IP version (IPv4 vs IPv6).

        Args:
            ip_prefix: Parent prefix to allocate from.
            target_prefix_length: Desired prefix length for the new prefix.
            at: Optional timestamp for point-in-time queries.

        Returns:
            The next available prefix, or None if no space is available.
        """
        if ip_prefix.version == 4:
            return await self._get_next_ipv4_prefix(
                ip_prefix=ip_prefix, target_prefix_length=target_prefix_length, at=at
            )
        return await self._get_next_ipv6_prefix(ip_prefix=ip_prefix, target_prefix_length=target_prefix_length, at=at)

    async def get_next_address(
        self, ip_prefix: IPNetworkType, at: Timestamp | str | None = None, is_pool: bool = False
    ) -> IPAddressType | None:
        """Get the next available free IP address within a prefix.

        Automatically selects the appropriate query based on IP version (IPv4 vs IPv6).

        Args:
            ip_prefix: Prefix to allocate an address from.
            at: Optional timestamp for point-in-time queries.
            is_pool: If True, includes network and broadcast addresses as allocatable.
                If False (default), reserves the first and last addresses.

        Returns:
            The next available IP address, or None if no addresses are available.
        """
        # Use IPv6-specific query for IPv6 to avoid 64-bit integer overflow
        query_class = IPv6PrefixIPAddressFetchFree if ip_prefix.version == 6 else IPPrefixIPAddressFetchFree
        query = await query_class.init(
            db=self.db,
            branch=self.branch,
            obj=ip_prefix,
            namespace=self.namespace,
            at=at,
            branch_agnostic=self.branch_agnostic,
            is_pool=is_pool,
        )
        await query.execute(db=self.db)
        return query.get_address()

    async def get_subnets(self, ip_prefix: IPNetworkType, at: Timestamp | str | None = None) -> Iterable[IPPrefixData]:
        """Get all subnets within a parent prefix.

        Args:
            ip_prefix: Parent prefix to query.
            at: Optional timestamp for point-in-time queries.

        Returns:
            Iterable of IPPrefixData objects representing child subnets.
        """
        query = await IPPrefixSubnetFetch.init(
            db=self.db,
            branch=self.branch,
            obj=ip_prefix,
            namespace=self.namespace,
            at=at,
            branch_agnostic=self.branch_agnostic,
        )
        await query.execute(db=self.db)
        return query.get_subnets()

    async def get_ip_addresses(
        self, ip_prefix: IPNetworkType, at: Timestamp | str | None = None
    ) -> Iterable[IPAddressData]:
        """Get all IP addresses within a prefix.

        Args:
            ip_prefix: Prefix to query.
            at: Optional timestamp for point-in-time queries.

        Returns:
            Iterable of IPAddressData objects representing addresses in the prefix.
        """
        query = await IPPrefixIPAddressFetch.init(
            db=self.db,
            branch=self.branch,
            obj=ip_prefix,
            namespace=self.namespace,
            at=at,
            branch_agnostic=self.branch_agnostic,
        )
        await query.execute(db=self.db)
        return query.get_addresses()
